# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Anything related to the help system that is common between interfaces"""

from .logging import make_logger
log = make_logger(__name__)

from collections import abc
import re

from . import (__appname__, __version__)
from .utils import expandtabs
from .utils.string import striplines
from .cliopts import DESCRIPTIONS as CLI_DESCRIPTIONS

MAIN_TOPICS = ('commands', 'settings', 'keymap', 'filters', 'rcfile')

ALIASES = {
    'cmds': 'commands',
    'config': 'settings', 'cfg': 'settings',
    'keys': 'keymap', 'keybindings': 'keymap',
    'filtering': 'filters',
    'rcfiles': 'rcfile',
}


import string
class ForgivingFormatter(string.Formatter):
    def get_value(self, key, args, kwargs):
        if isinstance(key, str):
            try:
                return kwargs[key]
            except KeyError:
                return '{%s}' % key
        else:
            return super().get_value(key, args, kwargs)

    def __call__(self, lines):
        return tuple(
            striplines(self.format(line, __appname__=__appname__)
                       for line in expandtabs.expand(lines, indent=2))
        )

finalize_lines = ForgivingFormatter()


class HelpManager():
    """Provide help texts for CommandManager, Settings and KeyMap objects"""
    def find(self, topic=None):
        """Find help topic and return lines"""
        if topic in ALIASES:
            topic = ALIASES[topic]

        if topic is None:
            return self.overview
        elif topic in MAIN_TOPICS:
            return getattr(self, topic)
        elif topic in self.cmdmgr:
            return self.command(topic)
        elif topic in self.localcfg or topic[4:] in self.remotecfg:
            return self.setting(topic)

        raise ValueError('Unknown help topic: %r' % topic)

    @property
    def overview(self):
        lines = [
            '{} {}'.format(__appname__, __version__),
            '',
            'SYNTAX',
            '\t{__appname__} [<OPTIONS>] [<COMMAND> ; <COMMAND> ; <COMMAND>...]',
            '',
        ]

        for section,opts in CLI_DESCRIPTIONS.items():
            lines.append('%s' % section.upper())
            for opts,desc in opts.items():
                lines.append('\t%s  \t%s' % (opts, desc))
            lines.append('')

        lines.append('For more information run:')
        for topic in MAIN_TOPICS:
            lines.append('\thelp %s' % topic)
        return finalize_lines(lines)

    @property
    def settings(self):
        """Return help text for all settings"""
        localcfg = self.localcfg
        remotecfg = self.remotecfg

        lines = [
            'SETTINGS',
            "\tSettings can be changed with the commands 'set' and 'reset'.",
            '',
            "\tUse an rc file (see 'help rcfile') to specify your custom defaults.",
            '',
        ]

        lines.append('\tLOCAL SETTINGS')
        for name in sorted(localcfg):
            lines.append('\t\t' + name + '  \t' + localcfg.description(name))

        lines += ['']

        lines.append('\tREMOTE SETTINGS')
        for name in sorted(remotecfg):
            lines.append('\t\tsrv.' + name + '  \t' + remotecfg.description(name))
        return finalize_lines(lines)

    def setting(self, name):
        """Return help text for setting"""
        if name in self.localcfg:
            cfg = self.localcfg
            key = name
        elif name[4:] in self.remotecfg:
            cfg = self.remotecfg
            key = name[4:]
        else:
            raise ValueError('Unknown help topic: %r' % name)
        value = cfg[key]

        def maybe_quote(value):
            if isinstance(value, str) and re.match(r'^\s+$', value):
                return repr(value)
            else:
                return str(value)

        lines = ['%s - \t%s' % (name, cfg.description(key)),
                 '\tValue: \t' + maybe_quote(cfg[key]),
                 '\tDefault: \t' + maybe_quote(cfg.default(key))]

        if hasattr(value, 'options'):
            opt_strs = []
            for opt in sorted(value.options):
                opt_strs.append(str(opt))
                aliases = [alias for alias,option in value.aliases.items()
                           if option == opt]
                if aliases:
                    opt_strs[-1] += ' (%s)' % (', '.join(aliases))
            lines.append('\tOptions: \t' + ', '.join(opt_strs))

        lines.append('\tSyntax: \t' + cfg.syntax(key))

        return finalize_lines(lines)

    @property
    def commands(self):
        """Must be set to a CommandManager object; provides a help text"""
        from .commands import (OPS_AND, OPS_OR, OPS_SEQ)
        lines = [
            'COMMANDS',
            '\tCommands can be called ',
            '\t\t- \tby providing them as command line arguments,',
            "\t\t- \tvia the internal command line (hit ':' to open it),",
            "\t\t- \tby binding them to keys (see 'help bind'),",
            ("\t\t- \tby listing them in an rc file (see 'help rcfile') "
             "and loading it with the '--rcfile' option or the 'rc' command."),
            '',
            '\tCHAINING COMMANDS',
            ("\t\tCombining commands with operators makes it possible to run "
             "a command based on the previous command's success."),
            "",
            "\t\tAvailable command operators are: ",
            "\t\t\t%s \t- \tRun the next command if the previous command succeeded." % '/'.join(OPS_AND),
            "\t\t\t%s \t- \tRun the next command if the previous command failed." % '/'.join(OPS_OR),
            "\t\t\t%s \t- \tRun the next command in any case." % '/'.join(OPS_SEQ),
            "",
            "\t\tCommand operators must be enclosed by spaces.",
            "",
            ("\t\tFor example, 'ls foo & ls bar' would list all 'foo' torrents and, "
             "if any where found, continue to list all 'bar' torrents.  "
             "However, 'ls foo | ls bar' would list 'bar' torrents only if there "
             "are no 'foo' torrents."),
            '',
            '\tGUESSING THE USER INTERFACE (CLI/TUI)',
            ("\t\tIf commands are given as command line arguments and neither "
             "'--tui' nor '--notui' are provided, {__appname__} tries to guess "
             "whether it makes sense to start the TUI or just run the commands "
             "and exit.  For example, if you run '{__appname__} stop foo', "
             "it is reasonable to assume that you want to run 'stop foo' and "
             "get your shell prompt back.  But if you run "
             "'{__appname__} set connect.host foo.bar', "
             "you probably expect the TUI to pop up."),
            '',
            "\t\tThis is how this works basically:",
            ("\t\t\t- \tWithout CLI commands, the TUI is loaded."),
            ("\t\t\t- \tCommands in the TORRENT category (see below) prevent the TUI."),
            ("\t\t\t- \tChanging TUI settings ('(re)set tui.*') enforces the TUI."),
            ("\t\t\t- \tChanging remote settings ('set srv.*') prevents the TUI."),
            ("\t\t\t- \tCommands that are exclusive to TUI or CLI "
             "(e.g. 'tab') enforce their interface.  Providing both TUI- "
             "and CLI-only commands produces an error."),
            '',
        ]

        cmdmgr = self.cmdmgr
        for category in cmdmgr.categories:
            lines.append('\t{} COMMANDS'.format(category.upper()))

            # To deduplicate commands with the same name that provide
            # different interfaces (but should have the same docs), map
            # command names to commands.
            cmds = {}
            for cmd in cmdmgr.all_commands:
                if category == cmd.category:
                    cmds[cmd.name] = cmd

            for cmdname,cmd in sorted(cmds.items()):
                lines.append('\t\t{}  \t{}'.format(', '.join((cmd.name,)+cmd.aliases),
                                                   cmd.description))
            lines.append('')
        return finalize_lines(lines)

    def command(self, name):
        """Return help text for command"""
        cmd = self.cmdmgr[name]

        def takes_value(argspec):
            if argspec.get('action') in ('store_true', 'store_false', 'store_const'):
                return False  # Boolean option
            if 'nargs' not in argspec:
                return True
            nargs = argspec['nargs']
            return not isinstance(nargs, int) or nargs > 0

        def arg_dest(argspec):
            if 'metavar' in argspec:
                dest = argspec['metavar'].upper()
            elif 'dest' in argspec:
                dest = argspec['dest'].upper()
            elif argspec['names'][0][0] == '-':
                dest = argspec['names'][0].lstrip('-').upper()
            else:
                return None

            if 'nargs' in argspec and argspec['nargs'] in ('*', '?'):
                return '[<%s>]' % dest
            else:
                return '<%s>' % dest

        lines = [cmd.name.upper()]

        log.debug('Generating help text for %s', cmd.name)
        names = ', '.join((cmd.name,) + cmd.aliases)
        lines = [names + ' - \t' + cmd.description]
        lines.append('')

        if cmd.usage:
            lines.append('USAGE')
            lines += ['\t' + u for u in cmd.usage]
            lines.append('')

        if cmd.argspecs:
            lines.append('ARGUMENTS')
            lines_args = []
            for argspec in cmd.argspecs:
                if 'description' not in argspec:
                    # Argument has no description
                    continue

                arglines = ['\t' + ', '.join(argspec['names'])]
                if takes_value(argspec):
                    dest = arg_dest(argspec)
                    if dest is not None:
                        arglines[0] += ' %s' % dest

                if isinstance(argspec['description'], str):
                    arglines[0] += '  \t' + argspec['description']
                else:  # Assume description is a sequence
                    arglines[0] += '  \t' + argspec['description'][0]
                    for paragraph in argspec['description'][1:]:
                        arglines.append('\t  \t' + paragraph)

                if 'document_default' not in argspec or argspec['document_default']:
                    # Argument takes a value that may default to another value
                    # if ommitted and we want to document that default value
                    def stringify_default(default):
                        dflt = default() if callable(default) else default

                        if not isinstance(dflt, str) and isinstance(dflt, abc.Sequence):
                            return ' '.join(dflt)
                        else:
                            return str(dflt)

                    if 'default_description' in argspec:
                        arglines.append('\t  \tDefault: %s' % stringify_default(argspec['default_description']))
                    elif 'default' in argspec:
                        arglines.append('\t  \tDefault: %s' % stringify_default(argspec['default']))

                lines_args.extend(arglines)
            lines += lines_args
            lines.append('')

        if cmd.examples:
            lines.append('EXAMPLES')
            lines += ['\t' + e for e in cmd.examples]
            lines.append('')

        for name,text in sorted(cmd.more_sections.items()):
            lines.append(name.upper())
            if callable(text):
                text = text()
            lines += ['\t' + line for line in text]
            lines.append('')

        return finalize_lines(lines)

    @property
    def keymap(self):
        """Must be set to a KeyMap object; provides a help text"""

        from .tui import main as tui
        km = tui.keymap
        lines = []

        def stringify(s):
            return ' '.join(s) if not isinstance(s, str) else s

        for context in sorted(km.contexts, key=lambda c: '' if c is None else c):
            if context is None:
                lines.append('GENERAL KEYBINDINGS')
            else:
                lines.append('{} KEYBINDINGS'.format(context.upper()))

            keymap = ((key, stringify(action)) for key,action in km.map(context))

            # Sort by command
            from natsort import (natsort_keygen, natsorted, ns)
            get_cmd = natsort_keygen(key=lambda pair: pair[1], alg=ns.IGNORECASE)
            for key,action in natsorted(keymap, key=get_cmd):
                if len(action) < 40:
                    lines.append('\t%s  \t%s  \t%s' % (key, action, km.get_description(key, context)))
                else:
                    lines.append('\t%s  \t%s' % (key, action))
                    lines.append('\t  \t%s' % (km.get_description(key, context),))
            lines.append('')
        return finalize_lines(lines)

    @property
    def filters(self):
        """Provide help text for arguments to TorrentFilter"""
        lines = [
            'FILTERING TORRENTS, FILES, PEERS, ETC',
            ('\tCommands that accept FILTER arguments are applied to items '
             'that match these filters.'),
            '',
            '\tThere are two kinds of filters:',
            '\t\t- \tBoolean filters stand on their own (e.g. "downloading")',
            '\t\t- \tComparative filters need a value (e.g. "seeds>20")',
            '',
            '\tThe syntax of comparative filters is: [[<FILTER NAME>]<OPERATOR>]<VALUE>',
            '',
            ('\tBesides the usual operators (=, !=, >, <, >=, <=), "~" matches '
             'if the item\'s value contains VALUE.'),
            '\tExample: "name~foo" matches all torrents with "foo" in their name.',
            '',
            ('\tIf FILTER NAME is omitted, it defaults to "name" for torrents and '
             'files and "domain" for trackers.  Peers don\'t have a default filter.  '
             'If OPERATOR is omitted, it defaults to "~".'),
            '\tExample: "foo" is the same as "~foo" is the same as "name~foo".',
            '',
            ('\tSpaces at the start and the end of VALUE are always removed.  '
             'If the result is enclosed by matching single or double quotes, they '
             'are removed.  Any other quotes are not interpreted, i.e. they must '
             'not be escaped.'),
            '\tExample: "name = foo " matches "foo"; "name = \' foo \' " matches " foo "',
            '',
            '\tAll filters can be inverted by prepending "!" to the filter name.',
            ('\tExample: "name!=foo" is the same as "!name=foo"; '
             '"!name!=foo" is the same as "name=foo".'),
            '',
            ('\tMatching strings is case-insensitive if VALUE does not contain any '
             'upper-case characters, otherwise it is case-sensitive.'),
            '',
            ('\tWhen matching numbers, the unit prefixes "k", "M", "G", "T" and '
             'their binary counterparts "Ki", "Mi", "Gi", "Ti" are supported.  '
             'The case of unit prefixes is ignored.'),
            ('\tExample: "size>1mi" is the same as "size>1048576" (1 Mebibyte); '
             '"size>1m" is the same as "size>1000000" (1 Megabyte)'),
            '',
            ('\tFor time-based filters, VALUE is either an absolute time stamp '
             'or a relative time delta based on the current time.'),
            '',
            ('\tTime stamps support a date in the format [[YYYY-]MM-]DD or YYYY[-MM] '
             'and a time in the format HH:MM[:SS].  Date and time can be combined by '
             'separating them with a space.'),
            '\tExamples: \t"added=2015-05" matches torrents that were added in May 2015.',
            '\t\t"completed>=01" matches torrents that finished downloading earlier this month.',
            ('\t\t"activity<10-17 18:45" matches torrents that were last active before '
             '18:45 (6:45 pm) on the 17th of October of this year.'),
            '',
            ('\tTime deltas use the format [in |+|-]AMOUNT[s|m|h|d|w|M|y][ ago].  '
             'The words "in" and "ago" are aliases for "+" and "-".  Negative time '
             'deltas match time stamps in the past and positive time deltas '
             'match time stamps in the future.  Filters have individual defaults for '
             'the sign; e.g. "eta > 1h" is the same as "eta > in 1h" while '
             '"completed > 1h" is the same as "completed > 1h ago".'),
            '',
            ('\tFilters can be combined with the operators "&" (logical AND) '
             'and "|" (logical OR).  Multiple FILTER arguments are implicitly '
             'combined with "|".'),
            '\tExample: "name=foo paused" is the same as "name=foo|paused".',
            '',
            ('\tOperators can be escaped with a preceding "\\" to remove their meaning.'),
            '\tExample: "name=foo\&bar" matches torrents with the name "foo&bar".',
        ]

        from .client.filters.torrent import SingleTorrentFilter
        from .client.filters.file import SingleTorrentFileFilter
        from .client.filters.peer import SingleTorrentPeerFilter
        from .client.filters.tracker import SingleTrackerFilter

        for caption,filt in (('TORRENT FILTERS', SingleTorrentFilter),
                             ('FILE FILTERS', SingleTorrentFileFilter),
                             ('PEER FILTERS', SingleTorrentPeerFilter),
                             ('TRACKER FILTERS', SingleTrackerFilter)):
            lines += ['', '\t%s' % caption]
            lines.append('\t\tBOOLEAN FILTERS')
            for fname,f in sorted(filt.BOOLEAN_FILTERS.items()):
                lines.append('\t\t\t{} \t{}'.format(', '.join((fname,)+f.aliases), f.description))
            lines += ['', '\t\tCOMPARATIVE FILTERS']
            for fname,f in sorted(filt.COMPARATIVE_FILTERS.items()):
                lines.append('\t\t\t{} \t{}'.format(', '.join((fname,)+f.aliases), f.description))

        return finalize_lines(lines)

    @property
    def rcfile(self):
        """Provide help text for rc file"""
        lines = [
            'RC FILES',
            ('\tAn rc file is a script that contains a list of arbitrary commands.  '
             'Commands can span multiple lines by escaping linebreaks with "\\".  '
             'Lines starting with "#" (optionally preceded by spaces) are ignored.'),
            '',
            ('\tThe default rc file path is "$XDG_CONFIG_HOME/{__appname__}/rc", '
             'where XDG_CONFIG_HOME defaults to "~/.config" if it is not set.'),
            '',
            ('\tA different path can be provided with the "--rcfile" option.  '
             'An existing rc file at the default path can be ignored with the '
             '"--norcfile" option.'),
            '',
            '\tTo permanently change the default config file, create an alias:',
            '',
            '\t\t$ alias stig="command stig --rcfile ~/.stigrc"',
            '',
            ('\tTo load an additional rc file after the default one, use the '
             '"rc" command.  (Note that this will prevent the TUI from being '
             'loaded unless you provide the "--tui" option.  See the GUESSING '
             'THE USER INTERFACE section in the "commands" help for more information.)'),
            '',
            ('\tCommands in an rc file are called during startup before the '
             'commands given on the command line.'),
            '',
            ('\tTUI commands (e.g. "tab" or "bind") in an rc file are ignored '
             'in CLI mode.'),
        ]
        return finalize_lines(lines)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering TrackerList items by various values"""

from ..ttypes import TorrentTracker
from . import (BoolFilterSpec, Filter, FilterChain, make_cmp_filter, CmpFilterSpec)
from . import (timestamp_or_timedelta, time_filter)


def _make_cmp_filter(*args, **kwargs):
    return make_cmp_filter(TorrentTracker.TYPES, *args, **kwargs)


class SingleTrackerFilter(Filter):
    DEFAULT_FILTER = 'domain'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda trk: True,
            aliases=('*',),
            description='All trackers'),
        'alive': BoolFilterSpec(
            lambda trk: trk['error'] == '' or trk['status'] == 'inactive',
            description='Trackers we are trying to connect to'),
    }

    COMPARATIVE_FILTERS = {
        'tier'           : _make_cmp_filter('tier',
                                            description='Match VALUE against torrent tier'),
        'domain'         : _make_cmp_filter('domain',
                                            aliases=('dom',),
                                            description='Match VALUE against domain from announce URL'),
        'url-announce'   : _make_cmp_filter('url-announce',
                                            aliases=('an',),
                                            description='Match VALUE against announce URL'),
        'url-scrape'     : _make_cmp_filter('url-scrape',
                                            aliases=('sc',),
                                            description='Match VALUE against scrape URL'),
        'status'         : _make_cmp_filter('status',
                                            aliases=('st',),
                                            description=('Match VALUE against tracker status '
                                                         '(stopped, idle, queued, announcing, scraping)')),
        'error'          : _make_cmp_filter('error',
                                            aliases=('err',),
                                            description='Match VALUE against error message from tracker'),
        'downloads'      : _make_cmp_filter('count-downloads',
                                            aliases=('dns',),
                                            description='Match VALUE against number of known downloads'),
        'leeches'        : _make_cmp_filter('count-leeches',
                                            aliases=('lcs',),
                                            description='Match VALUE against number of known downloading peers'),
        'seeds'          : _make_cmp_filter('count-seeds',
                                            aliases=('sds',),
                                            description='Match VALUE against number of known seeding peers'),
        'last-announce': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-last-announce'], op, v),
            aliases=('lan',),
            description='Match VALUE against time of last announce',
            needed_keys=('time-last-announce',),
            value_type=TorrentTracker.TYPES['time-last-announce'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
        'next-announce': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-next-announce'], op, v),
            aliases=('nan',),
            description='Match VALUE against time of next announce',
            needed_keys=('time-next-announce',),
            value_type=TorrentTracker.TYPES['time-next-announce'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=1),
        ),
        'last-scrape': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-last-scrape'], op, v),
            aliases=('lsc',),
            description='Match VALUE against time of last scrape',
            needed_keys=('time-last-scrape',),
            value_type=TorrentTracker.TYPES['time-last-scrape'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=-1),
        ),
        'next-scrape': CmpFilterSpec(
            lambda t, op, v: time_filter(t['time-next-scrape'], op, v),
            aliases=('nsc',),
            description='Match VALUE against time of next scrape',
            needed_keys=('time-next-scrape',),
            value_type=TorrentTracker.TYPES['time-next-scrape'],
            value_convert=lambda v: timestamp_or_timedelta(v, default_sign=1),
        ),
    }


class TorrentTrackerFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTrackerFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Display specifications for tables and such"""

from ..logging import make_logger
log = make_logger(__name__)

from collections import defaultdict
from time import time


class ColumnBase():
    header = {'left': '', 'right': ''}
    width = None
    min_width = 1
    align = 'right'
    wrap = 'clip'
    interfaces = ('cli', 'tui')
    may_have_wide_chars = False

    def __init__(self, data=None):
        self.data = data if data is not None else {}
        super().__init__()

    # _cache = defaultdict(lambda: {'value': None, 'last_hit': time(), 'hits': 0})
    _cache = defaultdict(lambda: {'value': None, 'last_hit': time()})
    _last_cache_prune = 0
    def _from_cache(self, create_value, *args):
        cache_id = (create_value, args)
        if cache_id not in self._cache:
            # log.debug('Calling %r with %r', create_value.__qualname__, args)
            value = self._cache[cache_id]['value'] = create_value(*args)
        else:
            cache_item = self._cache[cache_id]
            value = cache_item['value']
            cache_item['last_hit'] = time()
            # cache_item['hits'] += 1

            # Remove any cached items that haven't been used recently
            now = time()
            if now-ColumnBase._last_cache_prune > 60:
                ColumnBase._last_cache_prune = now

                # prune_counter = 0
                for cid,citem in tuple(self._cache.items()):
                    if now-citem['last_hit'] > 600:
                        del self._cache[cid]
                        # prune_counter += 1

                # # Debugging stuff; also uncomment the lines containing
                # # 'prune_counter' and 'hits'
                # items_without_hits = tuple(citem for citem in self._cache.values()
                #                            if citem['hits'] < 1)

                # log.debug('Cell cache: %r values pruned, %r remaining, '
                #           '%r with no hits, %r hits combined',
                #           prune_counter, len(self._cache), len(items_without_hits),
                #           sum((citem['hits'] for citem in self._cache.values())))


                # items_with_hits = tuple(citem for citem in self._cache.values()
                #                         if citem['hits'] > 0)
                # from time import (strftime, localtime)
                # for citem in items_with_hits:
                #     log.debug('  %5d hits, last hit: %s: %s',
                #               citem['hits'],
                #               strftime('%H:%M:%S', localtime(citem['last_hit'])),
                #               citem['value'])
        return value

    @classmethod
    def clearcache(cls):
        cls._cache.clear()

    def get_value(self):
        raise NotImplementedError()

    def get_cli_value(self):
        return self.get_value()

    def get_tui_value(self):
        return self.get_value()

    def get_raw_value(self):
        value = self.get_value()
        if isinstance(value, (int, float)) and value < float('inf'):
            return repr(value)
        else:
            return value

    def __repr__(self):
        if getattr(self, 'data', None):
            return '<{} {}>'.format(type(self).__name__, self.get_value())
        else:
            return '<{} <UNINITIALIZED>>'.format(type(self).__name__)

def _ensure_hide_unit(value):
    return type(value)(value, hide_unit=True)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from .errors import *
from .constants import *

from .utils import (Response, URL)

from .geoip import GeoIP
geoip = GeoIP()

from .sorters.torrent import TorrentSorter
from .sorters.peer import TorrentPeerSorter
from .sorters.tracker import TorrentTrackerSorter
from .sorters.setting import SettingSorter

from .filters.torrent import TorrentFilter
from .filters.file import TorrentFileFilter
from .filters.peer import TorrentPeerFilter
from .filters.tracker import TorrentTrackerFilter
from .filters.setting import SettingFilter

from .poll import RequestPoller
from .trequestpool import TorrentRequestPool

from .aiotransmission.rpc import TransmissionRPC
from .aiotransmission.api_status import StatusAPI
from .aiotransmission.api_settings import SettingsAPI
from .aiotransmission.api_torrent import TorrentAPI
from .api import API



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Documentation commands"""

from ...logging import make_logger
log = make_logger(__name__)

from ..base import misc as base
from .. import (ExpectedResource, CmdError)

class HelpCmd(base.HelpCmdbase):
    provides = {'cli'}
    srvapi = ExpectedResource

    async def run(self, TOPIC):
        # If TOPIC is a setting and it is managed by the server, we must fetch
        # config values from the server so we can display its current value.
        for topic in TOPIC:
            if topic.startswith('srv.'):
                try:
                    await self.srvapi.settings.update()
                except self.srvapi.ClientError as e:
                    self.error(e)
                finally:
                    break
        return super().run(TOPIC)

    def display_help(self, topics, lines):
        for line in lines:
            print(line)


class VersionCmd(base.VersionCmdbase):
    provides = {'cli'}


class LogCmd(base.LogCmdbase):
    provides = {'cli'}

    def _do(self, action, *args):
        cmd_str = '%s %s' % (action, ' '.join(args))
        raise CmdError('Unsupported command in CLI mode: %s' % cmd_str)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, CmdError, ExpectedResource)
from . import _mixin as mixin
from ._common import (make_X_FILTER_spec, make_COLUMNS_doc,
                      make_SORT_ORDERS_doc, make_SCRIPTING_doc)

import asyncio
from collections import abc


class ListTrackersCmdbase(mixin.get_tracker_sorter, mixin.get_tracker_columns,
                          mixin.get_tracker_filter, metaclass=InitCommand):
    name = 'trackerlist'
    aliases = ('trkls', 'lstrk')
    provides = set()
    category = 'tracker'
    description = 'List tracker(s) of torrent(s)'
    usage = ('trackerlist [<OPTIONS>]',
             'trackerlist [<OPTIONS>] <TORRENT FILTER>',
             'trackerlist [<OPTIONS>] <TORRENT FILTER> <TRACKER FILTER>')
    examples = ("trackerlist 'name~debian'",
                "trackerlist tracker=my.tracker.local error")
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),
        make_X_FILTER_spec('TRACKER', or_focused=True, nargs='?'),

        { 'names': ('--sort', '-s'),
          'default_description': "current value of 'sort.trackers' setting",
          'description': ('Comma-separated list of sort orders '
                          "(see SORT ORDERS section)") },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'columns.trackers' setting",
          'description': ('Comma-separated list of column names '
                          "(see COLUMNS section)") },
    )
    cfg = ExpectedResource

    from ...views.tracker import COLUMNS
    from ...client.sorters.tracker import TorrentTrackerSorter
    more_sections = {
        'COLUMNS': make_COLUMNS_doc(COLUMNS, '--columns', 'columns.trackers', append=(
            '',
            'The "torrent" column is added automatically if multiple '
            'torrents could be listed potentially.')),
        'SORT ORDERS': make_SORT_ORDERS_doc(TorrentTrackerSorter, '--sort', 'sort.trackers'),
        'SCRIPTING': make_SCRIPTING_doc(name),
    }

    async def run(self, TORRENT_FILTER, TRACKER_FILTER, sort, columns):
        columns = self.cfg['columns.trackers'] if columns is None else columns
        sort = self.cfg['sort.trackers'] if sort is None else sort
        try:
            torfilter = self.select_torrents(TORRENT_FILTER,
                                             allow_no_filter=True,
                                             discover_torrent=True)
            trkfilter = self.get_tracker_filter(TRACKER_FILTER)
            sort      = self.get_tracker_sorter(sort)
            columns   = self.get_tracker_columns(columns)
        except ValueError as e:
            raise CmdError(e)

        # Unless we're listing trackers of exactly one torrent, specified by its
        # ID, automatically add the 'torrent' column.
        if 'torrent' not in columns and \
           (not isinstance(torfilter, abc.Sequence) or len(torfilter) != 1):
            columns += ('torrent',)

        log.debug('Listing %s trackers of %s torrents', trkfilter, torfilter)

        if asyncio.iscoroutinefunction(self.make_tracker_list):
            await self.make_tracker_list(torfilter, trkfilter, sort, columns)
        else:
            self.make_tracker_list(torfilter, trkfilter, sort, columns)


class AnnounceCmdbase(metaclass=InitCommand):
    name = 'announce'
    aliases = ('an',)
    provides = set()
    category = 'tracker'
    description = 'Announce torrents to their trackers now if possible'
    usage = ('announce',
             'announce <TORRENT FILTER> <TORRENT FILTER> ...')
    examples = ('announce tracker~example.org',)
    argspecs = (
        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='*'),
    )
    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)
        else:
            response = await self.make_request(
                self.srvapi.torrent.announce(tfilter),
                polling_frenzy=False)
            if not response.success:
                raise CmdError()


class TrackerCmdbase(metaclass=InitCommand):
    name = 'tracker'
    aliases = ('trk',)
    provides = set()
    category = 'tracker'
    description = 'Add/Remove trackers to/from torrents'
    _ADD_ACTIONS = ('add',)
    _REMOVE_ACTIONS = ('remove', 'rm')
    _ALL_ACTIONS = _ADD_ACTIONS + _REMOVE_ACTIONS
    usage = ('tracker %s <URL>' % '|'.join(_ALL_ACTIONS),
             'tracker %s <TORRENT FILTER> <URL> <URL> ...' % '|'.join(_ALL_ACTIONS))

    examples = ('tracker add "torrent with no tracker" http://tracker3.example.org:12345/announce',
                'tracker remove all tracker1.example tracker2.example ')
    argspecs = (
        { 'names': ('ACTION',) },

        make_X_FILTER_spec('TORRENT', or_focused=True, nargs='?'),

        { 'names': ('URL',), 'nargs': '+',
          'description': ('Announce URL to add to or remove from matching torrents; '
                          'may be partial (e.g. domain name) when removing trackers') },
    )
    srvapi = ExpectedResource

    async def run(self, ACTION, TORRENT_FILTER, URL):
        urls = tuple(URL)
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            raise CmdError(e)

        if any(ACTION == action for action in self._ADD_ACTIONS):
            request = self.srvapi.torrent.tracker_add(tfilter, urls)
            log.debug('Adding trackers to %s torrents: %s', tfilter, ', '.join(urls))
        elif any(ACTION == action for action in self._REMOVE_ACTIONS):
            request = self.srvapi.torrent.tracker_remove(tfilter, urls, partial_match=True)
            log.debug('Removing trackers from %s torrents: %s', tfilter, ', '.join(urls))
        else:
            raise CmdError('Invalid ACTION: %r' % (ACTION,))

        response = await self.make_request(request, polling_frenzy=True)
        if not response.success:
            raise CmdError()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for torrent list columns"""

from ..logging import make_logger
log = make_logger(__name__)

from . import (ColumnBase, _ensure_hide_unit)

COLUMNS = {}
ALIASES = { 'mark'     : 'marked',
            'n'        : 'name',
            'dir'      : 'path',
            'st'       : 'status',
            'err'      : 'error',
            'up'       : 'uploaded',
            'dn'       : 'downloaded',
            '%dn'      : '%downloaded',
            'av'       : 'available',
            '%av'      : '%available',
            'sz'       : 'size',
            'prs'      : 'peers',
            'sds'      : 'seeds',
            'rto'      : 'ratio',
            'rup'      : 'rate-up',
            'rdn'      : 'rate-down',
            'lrup'     : 'limit-rate-up',
            'lrdn'     : 'limit-rate-down',
            'trk'      : 'tracker',
            'tcrt'     : 'created',
            'tadd'     : 'added',
            'tsta'     : 'started',
            'tact'     : 'activity',
            'tcmp'     : 'completed' }


class Marked(ColumnBase):
    interfaces = ('tui',)

COLUMNS['marked'] = Marked


class Id(ColumnBase):
    header = {'left': 'ID'}
    width = 4
    needed_keys = ('id',)
    align = 'right'

    def get_value(self):
        return self.data['id']

COLUMNS['id'] = Id


class Name(ColumnBase):
    header = {'left': 'Name'}
    width = None
    min_width = 5
    needed_keys = ('name',)
    align = 'left'
    may_have_wide_chars = True

    def get_value(self):
        return self.data['name']

COLUMNS['name'] = Name


import os
PATHSEP = os.sep
class Path(ColumnBase):
    header = {'left': 'Path'}
    width = None
    min_width = 10
    align = 'left'
    needed_keys = ('path',)
    may_have_wide_chars = True

    @staticmethod
    def _shorten_path(path, width):
        if width is None:
            return path
        elif width == 1:
            return '…'
        elif width <= 0:
            return ''

        # Remove characters from the end of the most toplevel directory until
        # there's only one left. Then do the same with the next level.
        #    /the/path/to/your/torrents
        #    /th/path/to/your/torrents
        #    /t/path/to/your/torrents
        #    /t/pat/to/your/torrents

        dirs = [d for d in path.split(PATHSEP) if d]
        calc_cur_len = lambda dirs: (sum(map(len, dirs)) +  # combined dir names
                                     len(dirs))             # path separators

        cur_len = calc_cur_len(dirs)
        while cur_len > width:
            for i,d in enumerate(dirs):
                if len(d) > 1:
                    dirs[i] = d[:-1]
                    break

            # Stop when every dir is one character long
            if cur_len <= len(dirs)*2:
                break

            cur_len = calc_cur_len(dirs)

        # Re-assemble shortened path
        path = PATHSEP + PATHSEP.join(dirs)

        # If "/t/p/t/y/t" is still too long, simply remove enough
        # characters from the front.
        if len(path) > width:
            excess = len(path) - width + 1
            path = '…' + path[excess:]

        return path

    def get_value(self):
        return self._from_cache(self._shorten_path,
                                self.data['path'].rstrip(PATHSEP),
                                self.width)

COLUMNS['path'] = Path


class Status(ColumnBase):
    header = {'left': 'Status'}
    width = 11
    min_width = 11
    needed_keys = ('status',)

    def get_value(self):
        return self.data['status'][0]

COLUMNS['status'] = Status


class Error(ColumnBase):
    header = {'left': 'Error'}
    width = ('weight', 300)
    min_width = 10
    needed_keys = ('error',)
    align = 'left'

    def get_value(self):
        return self.data['error']

COLUMNS['error'] = Error


class Uploaded(ColumnBase):
    header = {'left': 'Up', 'right': '?'}
    width = 6
    min_width = 6
    needed_keys = ('size-uploaded', 'size-downloaded')

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['size-uploaded'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['uploaded'] = Uploaded


class Downloaded(ColumnBase):
    header = {'left': 'Dn', 'right': '?'}
    width = 6
    min_width = 6
    needed_keys = ('size-downloaded', 'size-final')

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['size-downloaded'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['downloaded'] = Downloaded


class PercentDownloaded(ColumnBase):
    header = {'right': '%'}
    width = 4
    min_width = 4
    needed_keys = ('%verified', '%downloaded', '%metadata')

    @staticmethod
    def _get_value(verified, downloaded, metadata):
        if 0 < verified < 100:
            value = verified
        elif 0 < metadata < 100:
            value = metadata
        else:
            value = downloaded
        return _ensure_hide_unit(value)

    def get_value(self):
        t = self.data
        return self._from_cache(self._get_value, t['%verified'], t['%downloaded'], t['%metadata'])

COLUMNS['%downloaded'] = PercentDownloaded


class Available(ColumnBase):
    header = {'left': 'Avail', 'right': '?'}
    width = 7
    min_width = 7
    needed_keys = ('size-available', 'size-final', 'peers-seeding')

    @staticmethod
    def _get_value(size_final, size_available, peers_seeding):
        if peers_seeding > 0:
            return _ensure_hide_unit(size_final)
        else:
            return _ensure_hide_unit(size_available)

    def get_value(self):
        t = self.data
        return self._from_cache(self._get_value, t['size-final'], t['size-available'], t['peers-seeding'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['available'] = Available


class PercentAvailable(ColumnBase):
    header = {'left': 'Avail', 'right': '%'}
    width = 7
    min_width = 7
    needed_keys = ('%available', 'peers-seeding')

    @staticmethod
    def _get_value(perc_available, peers_seeding):
        if peers_seeding > 0:
            return type(perc_available)(100, unit='%', hide_unit=True)
        else:
            return _ensure_hide_unit(perc_available)

    def get_value(self):
        t = self.data
        return self._from_cache(self._get_value, t['%available'], t['peers-seeding'])

COLUMNS['%available'] = PercentAvailable


class Size(ColumnBase):
    header = {'left': 'Size', 'right': '?'}
    width = 6
    min_width = 6
    needed_keys = ('size-final',)

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['size-final'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['size'] = Size


class Peers(ColumnBase):
    header = {'left': 'Peers'}
    width = 5
    min_width = 5
    needed_keys = ('peers-connected',)

    def get_value(self):
        return self.data['peers-connected']

COLUMNS['peers'] = Peers


class Seeds(ColumnBase):
    header = {'left': 'Seeds'}
    width = 5
    min_width = 5
    needed_keys = ('peers-seeding',)

    def get_value(self):
        return self.data['peers-seeding']

COLUMNS['seeds'] = Seeds


class Ratio(ColumnBase):
    header = {'left': 'Ratio'}
    width = 5
    min_width = 5
    needed_keys = ('ratio',)

    def get_value(self):
        return self.data['ratio']

COLUMNS['ratio'] = Ratio


class RateUp(ColumnBase):
    header = {'left': 'Up', 'right': '?/s'}
    width = 6
    min_width = 6
    needed_keys = ('rate-up',)

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['rate-up'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-up'] = RateUp


class RateDown(ColumnBase):
    header = {'left': 'Dn', 'right': '?/s'}
    width = 6
    min_width = 6
    needed_keys = ('rate-down',)

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['rate-down'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-down'] = RateDown


class LimitRateUp(ColumnBase):
    header = {'left': 'LmtUp', 'right': '?/s'}
    width = 9
    min_width = 9
    needed_keys = ('limit-rate-up',)

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['limit-rate-up'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['limit-rate-up'] = LimitRateUp


class LimitRateDown(ColumnBase):
    header = {'left': 'LmtDn', 'right': '?/s'}
    width = 9
    min_width = 9
    needed_keys = ('limit-rate-down',)

    def get_value(self):
        return self._from_cache(_ensure_hide_unit, self.data['limit-rate-down'])

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['limit-rate-down'] = LimitRateDown


class Eta(ColumnBase):
    header = {'left': 'ETA'}
    width = 5
    min_width = 9
    needed_keys = ('timespan-eta',)

    def get_value(self):
        return self.data['timespan-eta']

COLUMNS['eta'] = Eta


class Tracker(ColumnBase):
    header = {'left': 'Tracker'}
    width = 10
    min_width = 5
    needed_keys = ('trackers',)
    align = 'left'

    def get_value(self):
        if len(self.data['trackers']) > 0:
            return self.data['trackers'][0]['url-announce'].domain
        else:
            return ''

COLUMNS['tracker'] = Tracker


class _TimeBase(ColumnBase):
    width = 10
    min_width = 10

class Created(_TimeBase):
    header = {'left': 'Created'}
    needed_keys = ('time-created',)

    def get_value(self):
        return self.data['time-created']

COLUMNS['created'] = Created

class Added(_TimeBase):
    header = {'left': 'Added'}
    needed_keys = ('time-added',)

    def get_value(self):
        return self.data['time-added']

COLUMNS['added'] = Added

class Started(_TimeBase):
    header = {'left': 'Started'}
    needed_keys = ('time-started',)

    def get_value(self):
        return self.data['time-started']

COLUMNS['started'] = Started

class Activity(_TimeBase):
    header = {'left': 'Activity'}
    needed_keys = ('time-activity',)

    def get_value(self):
        return self.data['time-activity']

COLUMNS['activity'] = Activity

class Completed(_TimeBase):
    header = {'left': 'Completed'}
    needed_keys = ('time-completed',)

    def get_value(self):
        return self.data['time-completed']

COLUMNS['completed'] = Completed



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Value types used in Torrent classes"""

# The TYPES dictionary at the end of this file maps Torrent key names to
# types.  Every Torrent key must have a type, even if it's just a no-op
# (`lambda obj: obj`).
#
# A type is any callable that converts a single value to the appropriate object.
#
# Types are used to convert values from the server (e.g. `Float(1234567)`) and
# for some types from the user as strings (e.g. `Float('1.3GB')`).

from ..logging import make_logger
log = make_logger(__name__)

from collections import abc
import os
import re
import time
import datetime

from .utils import (URL, Float, Int, convert, const)


Percent = Float.partial(unit='%')


class Ratio(Float):
    """A Torrent's upload/download ratio as a float"""
    INFINITE = float('inf')
    NOT_APPLICABLE = -1
    def __str__(self):
        if self == self.INFINITE:
            return '∞'
        elif self == self.NOT_APPLICABLE:
            return ''
        else:
            return super().without_unit


class Count(Int):
    UNKNOWN = -1
    def __str__(self):
        return '?' if self == self.UNKNOWN else super().without_unit


class Status(tuple):
    """A Torrent's status as a tuple of strings"""

    IDLE      = 'idle'
    DOWNLOAD  = 'downloading'
    UPLOAD    = 'uploading'
    CONNECTED = 'connected'
    SEED      = 'seeding'
    STOPPED   = 'stopped'
    QUEUED    = 'queued'
    ISOLATED  = 'isolated'
    VERIFY    = 'verifying'
    INIT      = 'discovering'
    ORDER = (VERIFY, DOWNLOAD, UPLOAD, INIT, CONNECTED,
             ISOLATED, QUEUED, IDLE, STOPPED, SEED)

    def __lt__(self, other):
        return self.ORDER.index(self[0]) < self.ORDER.index(other[0])

    def __le__(self, other):
        return self.ORDER.index(self[0]) <= self.ORDER.index(other[0])

    def __gt__(self, other):
        return self.ORDER.index(self[0]) > self.ORDER.index(other[0])

    def __ge__(self, other):
        return self.ORDER.index(self[0]) >= self.ORDER.index(other[0])



import operator
class SmartCmpStr(str):
    """
    String that compares case-insensitively if the other string consists solely
    of lower case characters.
    """

    def __cmp(self, op, other):
        if not isinstance(other, str):
            return NotImplemented

        # Do case-insensitive comparison if `other` consists solely of
        # lower-case characters
        o = str(other)
        if o == o.casefold():
            s = self.casefold()
        else:
            s = str(self)
        return op(s, o)

    def __lt__(self, other): return self.__cmp(operator.lt, other)
    def __le__(self, other): return self.__cmp(operator.le, other)
    def __eq__(self, other): return self.__cmp(operator.eq, other)
    def __ne__(self, other): return self.__cmp(operator.ne, other)
    def __gt__(self, other): return self.__cmp(operator.gt, other)
    def __ge__(self, other): return self.__cmp(operator.ge, other)
    def __contains__(self, other): return self.__cmp(operator.contains, other)

    # Defining __eq__ mandates defining __hash__ to make instances hashable
    def __hash__(self): return super().__hash__()


class Path(SmartCmpStr):
    def __new__(cls, path):
        return super().__new__(cls, os.path.normpath(path))

    def __repr__(self):
        return '<{} {!r}>'.format(type(self).__name__, str(self))

    def __hash__(self):
        return super().__hash__()



SECONDS = (('y', 31557600),  # 365.25 days
           ('M',  2629800),  # 1 year / 12
           ('w',   604800),  # 7 days
           ('d',    86400),
           ('h',     3600),
           ('m',       60),
           ('s',        1))

class Timedelta(int):
    # To sort unknown and not applicable Timedeltas below the rest, these
    # constants have large values that are very likely never encountered as
    # actual values.
    UNKNOWN        = 1e10
    NOT_APPLICABLE = 1e11

    _FULL_REGEX = re.compile((r'^(in |-|\+|)(\S+)( ago|)$'), flags=re.IGNORECASE)
    _SPLIT_REGEX = re.compile((r'((?:\d+\.\d+|\d+|\.\d+)[' +
                               r''.join(unit for unit,secs in SECONDS) +
                               r']?)'))

    @classmethod
    def from_string(cls, string):
        exc = ValueError('Invalid %s: %r' % (cls.__name__, string))

        match = cls._FULL_REGEX.match(string)
        if not match:
            raise exc

        sign_start = match.group(1).lower().strip()
        timespan = match.group(2)
        sign_end = match.group(3).lower().strip()
        sign = 1  # 1/-1 to represent +/-

        if sign_start == '-' or sign_end == 'ago':
            if sign_start == '+':
                raise exc
            sign = -1
        if any(sign_start == x for x in ('+', 'in')):
            if sign_end == 'ago':
                raise exc

        secs_total = 0
        for s in cls._SPLIT_REGEX.split(timespan):
            if len(s) < 1:
                continue
            elif not cls._SPLIT_REGEX.match(s):
                raise exc
            elif s[-1].isdigit():
                # No unit specified - assume seconds
                secs_total += float(s)
            else:
                num, unit = s[:-1], s[-1]
                for unit_,secs in SECONDS:
                    if unit == unit_:
                        secs_total += float(num) * secs
                        break

        return cls(secs_total * sign)

    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''
        elif self == 0:
            return '0s'

        abs_secs = abs(self)
        for i,(unit,amount) in enumerate(SECONDS):
            if abs_secs >= amount:
                num = self/amount

                # Small numbers get a sub-unit, for example '1d15h'
                if 1 <= abs_secs/amount < 10 and i < len(SECONDS)-1:
                    subunit, subamount = SECONDS[i+1]
                    if num >= 0:
                        subnum = abs(((num % 1) * amount) / subamount)
                    else:
                        subnum = abs(((num % -1) * amount) / subamount)

                    if subnum >= 1:
                        return '%d%s%d%s' % (int(num), unit, int(subnum), subunit)

                return '%d%s' % (int(num), unit)

    @property
    def with_preposition(self):
        if self > 0:
            return 'in %s' % self
        elif self < 0:
            return ('%s ago' % self)[1:]  # Remove the first char ('-')
        else:
            return 'now'

    def __bool__(self):
        """Whether delta is known"""
        return self not in (self.UNKNOWN, self.NOT_APPLICABLE)

    @property
    def is_known(self):
        return bool(self)

    @property
    def timestamp(self):
        if self == self.UNKNOWN:
            return Timestamp(Timestamp.UNKNOWN)
        elif self == self.NOT_APPLICABLE:
            return Timestamp(Timestamp.NOT_APPLICABLE)
        else:
            return Timestamp(self + time.time())

    def __repr__(self):
        return '<%s %s / %s>' % (type(self).__name__, super().__str__(), self.__str__())


class Timestamp(float):
    # These constants get "random" fractions added to make it less likely that
    # any real-world value equals them.
    NOW            = -2 + 0.123456789
    SOON           = -1 + 0.123456789
    UNKNOWN        = 1e10 + 0.123456789
    NOT_APPLICABLE = 1e11 + 0.123456789
    NEVER          = 1e12 + 0.123456789

    _FORMATS_DATE = (('%Y',       ('year',)),
                     ('%Y-%m',    ('year', 'month')),
                     ('%Y-%m-%d', ('year', 'month', 'day')),
                     ('%d',       ('day',)),
                     ('%m-%d',    ('month', 'day')))
    _FORMATS_TIME = (('%H:%M',    ('hour', 'minute')),
                     ('%H:%M:%S', ('hour', 'minute', 'second')))

    # Create all combinations of date, time and date+time formats, keeping track
    # of the values they specify
    _FORMATS = []
    for date_frmt,date_given in _FORMATS_DATE:
        _FORMATS.append((date_frmt, date_given))
        for time_frmt,time_given in _FORMATS_TIME:
            _FORMATS.append((time_frmt, time_given))
            given = date_given + time_given
            _FORMATS.append(('%s %s' % (date_frmt, time_frmt), given))
            _FORMATS.append(('%s %s' % (time_frmt, date_frmt), given))

    @classmethod
    def from_string(cls, string):
        string = string.strip().replace('  ', ' ')
        dt_now = datetime.datetime.now().replace(second=0, microsecond=0)
        dt_default = datetime.datetime.now().replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

        def fill_in_missing_values(dt, given):
            names = ('year', 'month', 'day', 'hour', 'minute', 'second')
            most_significant_given = names.index(given[0])
            new_args = {}
            for name in names:
                if name in given:
                    new_args[name] = getattr(dt, name)
                else:
                    # If `name` is less significant than the most significant
                    # user-given value, fill in value from current time,
                    # otherwise fill in the lowest possible value from default.
                    # Example: User gave month and day: Get year from t_now.
                    #          Get hours, minutes and seconds from t_default.
                    if most_significant_given > names.index(name):
                        new_args[name] = getattr(dt_now, name)
                    else:
                        new_args[name] = getattr(dt_default, name)
            return datetime.datetime(**new_args)

        def get_timespan(dt, given):
            # Return min/max timestamp based on the least significant value.
            # Example: Use gave '2012-08' (year and month): Return timestamp for
            #          '2012-08-01 00:00:00' and '2012-08-31 23:59:59'.
            least_significant_given = given[-1]
            if least_significant_given == 'year':
                ts_min = datetime.datetime(year=dt.year, month=1, day=1)
                ts_max = datetime.datetime(year=dt.year+1, month=1, day=1) - datetime.timedelta(seconds=1)
            elif least_significant_given == 'month':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=1)
                if dt.month < 12:
                    ts_max = (datetime.datetime(year=dt.year, month=dt.month+1, day=1)
                              - datetime.timedelta(seconds=1))
                else:
                    ts_max = (datetime.datetime(year=dt.year+1, month=1, day=1)
                              - datetime.timedelta(seconds=1))
            elif least_significant_given == 'day':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=dt.day)
                ts_max = ts_min + datetime.timedelta(hours=23, minutes=59, seconds=59)
            elif least_significant_given == 'hour':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, hour=dt.hour)
                ts_max = ts_min + datetime.timedelta(minutes=59, seconds=59)
            elif least_significant_given == 'minute':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, hour=dt.hour, minute=dt.minute)
                ts_max = ts_min + datetime.timedelta(seconds=59)
            else:
                ts_min, ts_max = dt, dt
            return (int(ts_min.timestamp()), int(ts_max.timestamp()))

        for frmt,given in cls._FORMATS:
            try:
                dt = datetime.datetime.strptime(string, frmt)
            except ValueError:
                continue
            else:
                dt = fill_in_missing_values(dt, given)
                return cls(dt.timestamp(), _timespan=get_timespan(dt, given))

        raise ValueError('Invalid format: %r' % string)

    def __new__(cls, seconds, _timespan=None):
        obj = super().__new__(cls, seconds)
        obj._timespan = _timespan if _timespan is not None else (seconds, seconds)
        return obj

    def __eq__(self, other):
        return self._timespan[0] <= other <= self._timespan[1]

    def __hash__(self):  # Needed to stay hashable with __eq__()
        return super().__hash__()

    def __gt__(self, other):
        return self._timespan[0] > other

    def __ge__(self, other):
        return self._timespan[1] >= other

    def __lt__(self, other):
        return self._timespan[1] < other

    def __le__(self, other):
        return self._timespan[0] <= other

    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''
        elif self == self.NOW:
            return 'now'
        elif self == self.SOON:
            return 'soon'
        elif self == self.NEVER:
            return 'never'

        # The format is based on the delta of seconds compared to the current time
        abs_delta = abs(self - time.time())
        if abs_delta < 120:
            frmt = '%H:%M:%S'
        elif abs_delta < 86400:
            frmt = '%H:%M'
        else:
            frmt = '%Y-%m-%d'
        return time.strftime(frmt, time.localtime(self))

    @property
    def full(self):
        if self == self.UNKNOWN:
            return 'unknown'
        elif self == self.NOT_APPLICABLE:
            return 'not applicable'
        elif self == self.NOW:
            return 'now'
        elif self == self.SOON:
            return 'soon'
        elif self == self.NEVER:
            return 'never'
        else:
            return time.strftime('%Y-%m-%d %H:%M', time.localtime(self))

    def __bool__(self):
        """Whether timestamp known"""
        return self not in (self.UNKNOWN, self.NOT_APPLICABLE, self.NEVER)

    @property
    def is_known(self):
        return self not in (self.NOW, self.SOON, self.UNKNOWN, self.NOT_APPLICABLE, self.NEVER)

    @property
    def timedelta(self):
        if self == self.UNKNOWN:
            return Timedelta(Timedelta.UNKNOWN)
        elif self == self.NOT_APPLICABLE:
            return Timedelta(Timedelta.NOT_APPLICABLE)
        elif self == self.NOW:
            return Timedelta(0)
        elif self == self.SOON:
            return Timedelta(0)
        elif self == self.NEVER:
            return Timedelta(Timedelta.NOT_APPLICABLE)
        else:
            return Timedelta(self - time.time())

    @property
    def in_future(self):
        return bool(self) and self > time.time()

    def __repr__(self):
        return '<%s %s / %s>' % (type(self).__name__, super().__str__(), self.__str__())



class TorrentFilePriority(str):
    INT2STR = {-2:'off', -1:'low', 0:'normal', 1:'high'}
    STR2INT = {'off':-2, 'low':-1, 'normal':0, 'high':1}

    def __new__(cls, prio):
        if isinstance(prio, int):
            if prio not in cls.INT2STR:
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, prio))
            obj = super().__new__(cls, cls.INT2STR[prio])
        else:
            if prio not in cls.STR2INT:
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, prio))
            obj = super().__new__(cls, prio)
        return obj

    def __lt__(self, other): return int(self) < int(other)
    def __gt__(self, other): return int(self) > int(other)
    def __le__(self, other): return int(self) <= int(other)
    def __ge__(self, other): return int(self) >= int(other)
    def __int__(self): return self.STR2INT[self]
    def __repr__(self): return '<%s %r>' % (type(self).__name__, str(self))

    def __eq__(self, other):
        if isinstance(other, int):
            return super().__eq__(self.INT2STR.get(other))
        return super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super().__hash__()


class TorrentFile(abc.Mapping):
    """Mapping that holds the values of a single file in a torrent"""

    # Distinguish subtrees from files without comparing classes everywhere
    nodetype = 'leaf'

    TYPES = {
        'id'              : lambda val: val,
        'tid'             : lambda val: val,
        'name'            : SmartCmpStr,
        'path-absolute'   : Path,
        'path-relative'   : Path,
        'location'        : Path,
        'size-total'      : lambda size: convert.size(size, unit='byte'),
        'size-downloaded' : lambda size: convert.size(size, unit='byte'),
        'is-wanted'       : bool,
        'priority'        : TorrentFilePriority,
        '%downloaded'     : Percent,
    }

    _MODIFIERS = {
        'id'              : lambda raw: raw['id'],
        'tid'             : lambda raw: raw['tid'],
        'name'            : lambda raw: raw['name'],
        'path-absolute'   : lambda raw: os.path.join(raw['location'], raw['path'], raw['name']),
        'path-relative'   : lambda raw: os.path.join(raw['path'], raw['name']),
        'location'        : lambda raw: raw['location'],
        'size-total'      : lambda raw: raw['size-total'],
        'size-downloaded' : lambda raw: raw['size-downloaded'],
        'is-wanted'       : lambda raw: raw['is-wanted'],
        'priority'        : lambda raw: 'off' if not raw['is-wanted'] else raw['priority'],
        '%downloaded'     : lambda raw: _calc_percent(raw['size-downloaded'], raw['size-total']),
    }

    def __init__(self, tid, id, name, path, location, size_total, size_downloaded, is_wanted, priority):
        self._raw = {'tid': tid, 'id': id, 'name': name, 'path': path, 'location': location,
                     'is-wanted': is_wanted, 'priority': priority,
                     'size-total': size_total, 'size-downloaded': size_downloaded}
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            val = self._MODIFIERS[key](self._raw)
            typ = self.TYPES.get(key)
            if typ is not None:
                self._cache[key] = typ(val)
            else:
                self._cache[key] = val
        return self._cache[key]

    def update(self, raw):
        self._raw.update(raw)
        cache = self._cache
        for key,new_value in raw.items():
            cached_value = cache.get(key)
            if cached_value is not None and cached_value != new_value:
                del cache[key]

        # %downloaded is never in raw because it is calculated from
        # size-downloaded and size-total
        if 'size-downloaded' in raw:
            try:
                del cache['%downloaded']
            except KeyError:
                pass

    def __repr__(self): return '<{} {!r}>'.format(type(self).__name__, self['name'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)


from . import base
def _ensure_TorrentFileTree(obj):
    if isinstance(obj, base.TorrentFileTreeBase):
        return obj
    else:
        raise RuntimeError('Not a TorrentFileTreeBase: %r' % obj)




from collections import (defaultdict, deque)
MAX_SAMPLES = 10
MAX_SAMPLE_AGE = 5*3600
_PEER_PROGRESS_DATA = defaultdict(lambda: deque(maxlen=MAX_SAMPLES))

def gc_peer_progress_data():
    for peer_id,samples in tuple(_PEER_PROGRESS_DATA.items()):
        # Remove samples that are too old
        while samples and (samples[0][0] + MAX_SAMPLE_AGE) < time.monotonic():
            samples.popleft()

        # Remove deque if there are no samples left
        if not samples:
            del _PEER_PROGRESS_DATA[peer_id]

def _guess_peer_rate_and_eta(peer_id, peer_progress, torrent_size):
    rate = 0
    if peer_progress >= 1:
        # Peer has already downloaded everything
        eta = Timedelta.NOT_APPLICABLE
    else:
        eta = Timedelta.UNKNOWN
        samples = _PEER_PROGRESS_DATA[peer_id]

        # Don't add the same progress twice
        if not samples or peer_progress != samples[-1][1]:
            samples.append((time.monotonic(), peer_progress))

        # We need at least 3 samples
        if len(samples) >= 3:
            # Use second and last sample to calculate rate.  The first sample is
            # ignored because its timestamp may be inaccurate: When we add the
            # first sample, the peer's progress is not current but the latest we
            # received, which happened likely tens of seconds ago.
            t_first, p_first = samples[1]
            t_last, p_last = samples[-1]
            p_diff = p_last - p_first
            t_diff = t_last - t_first

            # It's possible progress goes down, e.g. if a peer deletes a file
            if p_diff > 0:
                torrent_size = int(torrent_size)  # Don't copy unit + unit prefix from torrent_size
                size_diff = torrent_size * p_diff
                rate = size_diff / t_diff
                size_remaining = torrent_size - (torrent_size * peer_progress)
                eta = size_remaining / rate

    return rate, eta

from . import geoip
class TorrentPeer(abc.Mapping):
    TYPES = {
        'id'          : lambda val: val,
        'tid'         : lambda val: val,
        'tname'       : SmartCmpStr,
        'tsize'       : lambda size: convert.size(size, unit='byte'),
        'ip'          : str,
        'port'        : int,
        'client'      : SmartCmpStr,
        'country'     : SmartCmpStr,
        '%downloaded' : Percent,
        'rate-up'     : lambda rate: convert.bandwidth(rate, unit='byte'),
        'rate-down'   : lambda rate: convert.bandwidth(rate, unit='byte'),
        'eta'         : Timedelta,
        'rate-est'    : lambda rate: convert.bandwidth(rate, unit='byte'),
    }

    _MODIFIERS = {
        'id'      : lambda p: (p['tid'], p['ip'], p['port']),
        'country' : lambda p: geoip.country_code(p['ip']) or '?',
    }

    def __init__(self, tid, tname, tsize, ip, port, client, pdownloaded, rate_up, rate_down):
        self._dct = {'tid': tid, 'tname': tname, 'tsize': tsize,
                     'ip': ip, 'port': port, 'client': client, '%downloaded': pdownloaded,
                     'rate-up': rate_up, 'rate-down': rate_down}
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            if key in ('eta', 'rate-est'):
                rate, eta = _guess_peer_rate_and_eta(self['id'], self['%downloaded'] / 100, self['tsize'])
                self._cache['rate-est'] = self.TYPES['rate-est'](rate)
                self._cache['eta'] = self.TYPES['eta'](eta)

            else:
                if key in self._MODIFIERS:
                    val = self._MODIFIERS[key](self._dct)
                else:
                    val = self._dct[key]
                self._cache[key] = self.TYPES[key](val)
        return self._cache[key]

    def clearcache(self):
        self._cache.clear()

    def __repr__(self): return '<{} #{}, {}>'.format(type(self).__name__, self['tid'], self['ip'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)



class TorrentTracker(abc.Mapping):
    def _validate_tracker_status(string):
        if string not in ('stopped', 'idle', 'queued', 'announcing', 'scraping'):
            raise ValueError('Invalid tracker status: %r' % string)
        else:
            return SmartCmpStr(string)

    TYPES = {
        'id'                 : lambda val: val,
        'tid'                : int,
        'tname'              : SmartCmpStr,
        'tier'               : int,

        'url-announce'       : URL,
        'url-scrape'         : URL,
        'domain'             : SmartCmpStr,

        'status-announce'    : _validate_tracker_status,
        'status-scrape'      : _validate_tracker_status,
        'status'             : _validate_tracker_status,

        'error-announce'     : SmartCmpStr,
        'error-scrape'       : SmartCmpStr,
        'error'              : SmartCmpStr,

        'count-downloads'    : Count,
        'count-leeches'      : Count,
        'count-seeds'        : Count,

        'time-last-announce' : Timestamp,
        'time-next-announce' : Timestamp,
        'time-last-scrape'   : Timestamp,
        'time-next-scrape'   : Timestamp,
    }

    _MODIFIERS = {
        'id'     : lambda self: hash((self['tid'], self['url-announce'])),
        'domain' : lambda self: self['url-announce'].domain,
        'status' : lambda self: (self['status-scrape'] if self['status-announce'] == 'idle'
                                 else self['status-announce']),
        'error'  : lambda self: ('Announce error: %s' % self['error-announce']
                                 if self['error-announce'] else
                                 'Scrape error: %s' % self['error-scrape']
                                 if self['error-scrape'] else '')
    }

    def __init__(self, trkdict):
        self._dct = trkdict
        self._cache = {}

    def __getitem__(self, key):
        cache = self._cache
        if key not in self._cache:
            modifier = self._MODIFIERS.get(key)
            if modifier is not None:
                val = modifier(self)
            else:
                val = self._dct[key]
            cache[key] = self.TYPES[key](val)

        return cache[key]

    def __repr__(self): return '<%s %s>' % (type(self).__name__, self['url-announce'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)



def _rate(rate):
    return convert.bandwidth(rate, unit='byte')


def _rate_limit(limit):
    return const.UNLIMITED if limit is None else convert.bandwidth(limit, unit='byte')


def _data_size(size):
    return convert.size(size, unit='byte')


def _calc_percent(a, b):
    try:
        return a / b * 100
    except ZeroDivisionError:
        return 0


TYPES = {
    'id'                           : int,
    'hash'                         : str,
    'name'                         : SmartCmpStr,
    'ratio'                        : Ratio,
    'status'                       : Status,
    'path'                         : Path,
    'private'                      : bool,
    'comment'                      : SmartCmpStr,
    'creator'                      : SmartCmpStr,
    'magnetlink'                   : str,
    'count-pieces'                 : Int,

    '%downloaded'                  : Percent,
    '%uploaded'                    : Percent,
    '%metadata'                    : Percent,
    '%verified'                    : Percent,
    '%available'                   : Percent,

    'peers-connected'              : Int,
    'peers-uploading'              : Int,
    'peers-downloading'            : Int,
    'peers-seeding'                : Count,

    'timespan-eta'                 : Timedelta,
    'timespan-seeding'             : Timedelta,
    'timespan-downloading'         : Timedelta,
    'time-created'                 : Timestamp,
    'time-added'                   : Timestamp,
    'time-started'                 : Timestamp,
    'time-activity'                : Timestamp,
    'time-completed'               : Timestamp,
    'time-manual-announce-allowed' : Timestamp,

    'rate-down'                    : _rate,
    'rate-up'                      : _rate,

    'limit-rate-down'              : _rate_limit,
    'limit-rate-up'                : _rate_limit,

    'size-final'                   : _data_size,
    'size-total'                   : _data_size,
    'size-downloaded'              : _data_size,
    'size-uploaded'                : _data_size,
    'size-available'               : _data_size,
    'size-left'                    : _data_size,
    'size-corrupt'                 : _data_size,
    'size-piece'                   : _data_size,

    'error'                        : str,
    'trackers'                     : tuple,
    'peers'                        : tuple,
    'files'                        : _ensure_TorrentFileTree,
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from . import (SortSpecBase, SorterBase)


class _SortSpec(SortSpecBase):
    def __init__(self, *args, description='', **kwargs):
        description = 'Sort settings by %s' % description
        super().__init__(*args, description=description, **kwargs)


class SettingSorter(SorterBase):
    SORTSPECS = {
        'name'        : _SortSpec(lambda s: s['id'],
                                  description='name'),
        # repr() because values have incompatible types
        'value'       : _SortSpec(lambda s: repr(s['value']),
                                  description='value'),
        'default'     : _SortSpec(lambda s: repr(s['default']),
                                  description='default'),
        'description' : _SortSpec(lambda s: s['description'],
                                  description='description'),
    }
    DEFAULT_SORT = 'name'



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from string import hexdigits as HEXDIGITS
from collections import abc
import os
import base64

from ..utils import (Response, URL)
from .torrent import (TorrentFields, Torrent)
from .. import ClientError
from ..filters.torrent import TorrentFilter
from ..filters.file import TorrentFileFilter
from ..utils import (Bool, Bandwidth, BoolOrBandwidth)
from ..ttypes import Path


class _TorrentCache():
    def __init__(self, raw_torrents=()):
        self._tdict = {}  # Map torrent IDs to Torrent objects

    def update(self, raw_torrents):
        # import time ; start = time.time()
        tdict = self._tdict
        for rt in raw_torrents:
            tid = rt['id']
            if tid in tdict:
                # Update existing torrent
                # log.debug('Updating torrent #%d, %d keys: %s', tid, len(rt), tuple(rt))
                tdict[tid].update(rt)
            else:
                # Add new torrent
                # log.debug('Adding torrent #%d, %d keys: %s', tid, len(rt), tuple(rt))
                tdict[tid] = Torrent(rt)
        # log.debug('Updated %d cached with %d new torrents in %.3fms',
        #           len(tdict), len(raw_torrents), (time.time()-start)*1000)

    def purge(self, existing_tids):
        """Remove torrents with IDs that are not in `existing_ids`"""
        tdict = self._tdict
        known_tids = set(tdict)
        removed_tids = known_tids.difference(existing_tids)
        if removed_tids:
            log.debug('Clearing cached torrents: %r', removed_tids)
        for tid in removed_tids:
            del tdict[tid]

    def get(self, *ids):
        """Return tuple of Torrent objects"""
        if ids:
            return tuple(t for tid,t in self._tdict.items() if tid in ids)
        else:
            return tuple(self._tdict.values())

    def __len__(self):
        return len(self._tdict)

    def __repr__(self):
        tlist = ', '.join('#'+str(tid) for tid in self._tdict)
        return '<{cls} {tlist}>'.format(cls=type(self).__name__,
                                        tlist=tlist or '(empty)')


class TorrentAPI():
    """High-level abstraction of the Transmission RPC protocol"""

    def __init__(self, rpc):
        self.rpc = rpc
        self._tcache = _TorrentCache()

    def clearcache(self):
        """Remove all torrents from cache"""
        self._tcache.purge(existing_tids=())

    @staticmethod
    async def _request(method, *args, **kwargs):
        try:
            result = await method(*args, **kwargs)
        except ClientError as e:
            return Response(success=False, result=None, errors=(str(e),))
        else:
            return Response(success=True, result=result)

    async def _map_tid_to_torrent_values(self, torrents, keys):
        """
        Map torrent ID to Torrent value(s)

        If `keys` lists only one key, the returned map maps each torrent's ID to
        its value of that key.

        If `keys` lists two or more keys, the returned map maps torrent IDs to a
        dict that maps keys to each torrent's value for that key.

        The result is returned attached to a Response instance via the attribute
        'torrent_values'.
        """
        response = await self.torrents(torrents, keys=('id',) + tuple(keys))
        if not response.success:
            return Response(success=False, torrent_values={}, errors=response.errors)
        else:
            if len(keys) == 1:
                key = keys[0]
                torrent_values = {t['id']:t[key] for t in response.torrents}
            else:
                torrent_values = {t['id']:{key:t[key] for key in keys}
                                  for t in response.torrents}
            return Response(success=True, torrent_values=torrent_values)

    async def _abs_download_path(self, path):
        """Turn relative `path` into absolute path based on default download path"""
        if not os.path.isabs(path):
            response = await self._request(self.rpc.session_get)
            if not response.success:
                return Response(success=False, path=None, errors=response.errors)
            else:
                download_dir = response.result['download-dir']
                abs_path = os.path.normpath(os.path.join(download_dir, path))
                return Response(success=True, path=Path(abs_path))
        return Response(success=True, path=Path(path))

    async def add(self, torrent, stopped=False, path=None):
        """
        Add torrent from file, URL or hash

        torrent: Path to local file, web/magnet link or hash
        stopped: False to start downloading immediately, True otherwise
        path:    Download directory or `None` for default directory

        Return Response with the following properties:
            torrent: Torrent object with the keys 'id' and 'name' if the
                     torrent could be added or if it already exists, otherwise
                     None
            success: False if torrent could not be added or already exists,
                     True otherwise
            msgs:    List of info messages
            errors:  List of error messages
        """
        torrent_str = torrent
        args = {'paused': bool(stopped)}

        if path is not None:
            response = await self._abs_download_path(path)
            if not response.success:
                return Response(success=False, torrent=None, errors=response.errors)
            else:
                args['download-dir'] = response.path

        # Check if torrent_str is path to local torrent file
        torrent_str_path = os.path.expanduser(torrent_str)
        if os.path.exists(torrent_str_path):
            torrent_str = torrent_str_path
            del torrent_str_path

            # Read local file
            try:
                with open(torrent_str, 'rb') as f:
                    args['metainfo'] = str(base64.b64encode(f.read()),
                                           encoding='ascii')
            except OSError as e:
                return Response(success=False, torrent=None,
                                errors=('%s: %s' % (e.strerror, torrent_str),))
        elif len(torrent_str) == 40 and all(c in HEXDIGITS for c in torrent_str):
            # Convert hash to magnet link
            args['filename'] = 'magnet:?xt=urn:btih:' + torrent_str
        else:
            # It's either a link or a torrent file on the server - let the
            # daemon figure it out
            args['filename'] = torrent_str

        response = await self._request(self.rpc.torrent_add, **args)
        if not response.success:
            errors = []
            for error in response.errors:
                if 'Invalid or corrupt' in error:
                    errors.append('Torrent file is corrupt or doesn\'t exist: %r' % torrent_str)
                else:
                    errors.append(error)
            return Response(success=False, torrent=None, msgs=response.msgs, errors=errors)
        else:
            result = response.result
            errors = ()
            msgs = ()
            if 'torrent-duplicate' in result:
                info = result['torrent-duplicate']
                errors = ('Torrent already exists: ' + info['name'],)
                success = False
            elif 'torrent-added' in result:
                info = result['torrent-added']
                msgs = ('Added %s' % info['name'],)
                success = True
            else:
                raise RuntimeError('Malformed response: %r' % (result,))
            torrent = Torrent({'id': info['id'], 'name': info['name']})
            return Response(success=success, torrent=torrent, msgs=msgs, errors=errors)


    async def _request_torrents(self, fields, ids=None):
        """
        Make 'torrent-get' RPC request

        Return a Response object with 'raw_torrents' set to a tuple of torrents
        according to the RPC spec.
        """
        if 'id' not in fields:
            fields = ('id',) + tuple(fields)
        try:
            if ids is None:
                # Request all IDs
                raw_tlist = await self.rpc.torrent_get(fields=fields)
            else:
                if len(ids) > 0:
                    # Request given IDs
                    raw_tlist = await self.rpc.torrent_get(fields=fields, ids=ids)
                else:
                    # No IDs (i.e. empty torrent list) requested
                    raw_tlist = []
        except ClientError as e:
            return Response(success=False, raw_torrents=(), errors=(str(e),))
        else:
            self._tcache.update(raw_tlist)

            # If we just got a list of all torrents, we can check for torrents
            # that we still have cached but don't exist anymore and purge them.
            if ids is None:
                tids = tuple(t['id'] for t in raw_tlist)
                self._tcache.purge(existing_tids=tids)

            return Response(success=True, raw_torrents=raw_tlist)

    async def _get_torrents_by_ids(self, keys, ids=None):
        """
        Return a Response object with 'torrents' set to a tuple of Torrents

        keys: 'ALL' for all supported Torrent keys or a sequence of key
              strings (see client.ttypes.TYPES for available keys)
        ids:  None for all torrents or a sequence of wanted IDs
        """
        if keys == 'ALL':
            fields = TorrentFields(keys)
        else:
            fields = TorrentFields(*keys)

        response = await self._request_torrents(fields, ids)
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            from time import time
            start = time()

            errors = []

            # Get torrents from cache
            if ids is None:
                tlist = self._tcache.get()  # All torrents
            elif not ids:
                tlist = ()  # No torrents requested
            else:
                tlist = self._tcache.get(*ids)

                # Provide error for requested IDs that don't exist
                existing_ids = tuple(t['id'] for t in tlist)
                for tid in ids:
                    # Torrent objects are equal to an integer of the torrent's ID
                    if tid not in existing_ids:
                        errors.append('No torrent with ID: %d' % tid)

            success = len(tlist) > 0 or not ids
            log.debug('Found %d torrents in %.3fms', len(tlist), (time()-start)*1e3)
        return Response(success=success, torrents=tlist, errors=errors)

    async def _get_torrents_by_filter(self, keys, tfilter=None):
        """
        Return a Response object with 'torrents' set to a tuple of Torrents

        keys:    See _get_torrents_by_ids
        tfilter: A TorrentFilter instance or None to get all torrents
        """
        if tfilter is None:
            log.debug('Looking for all torrents with keys: %s', keys)
            # No filter specified - just return all torrents with the specified keys
            return await self._get_torrents_by_ids(keys=keys)
        else:
            log.debug('Looking for %s torrents with keys: %s', tfilter, keys)
            if isinstance(tfilter, str):
                tfilter = TorrentFilter(tfilter)

            tlist = ()

            # Request all torrents with the keys needed to filter them
            log.debug('Requesting full list with filter keys: %s', tfilter.needed_keys)
            response = await self._get_torrents_by_ids(keys=tfilter.needed_keys)
            if not response.success:
                return Response(success=False, torrents=(), errors=response.errors)
            else:
                # Find IDs of torrents that match tfilter
                wanted_ids = tuple(t['id'] for t in tfilter.apply(response.torrents))
                log.debug('Wanted IDs: %s', wanted_ids)
                if len(wanted_ids) > 0:
                    # Get only wanted torrents with all wanted keys
                    response = await self._get_torrents_by_ids(keys, wanted_ids)
                    if not response.success:
                        return Response(success=False, torrents=(), errors=response.errors)
                    else:
                        tlist = tuple(response.torrents)

            success = len(tlist) > 0
            msgs = errors = ()
            if not success:
                errors = ('No matching torrents: %s' % (tfilter,),)
            else:
                msgs = ('Found %d %s torrent%s' %
                        (len(tlist), tfilter, '' if len(tlist) == 1 else 's'),)
            return Response(success=success, torrents=tlist, msgs=msgs, errors=errors)

    async def torrents(self, torrents=None, keys='ALL'):
        """
        Get torrents

        torrents: Iterator of torrent IDs, TorrentFilter object (or its string
                  representation) or None for all torrents
        keys: tuple of Torrent keys to fetch or 'ALL' for all torrents

        Return Response with the following properties:
            torrents: Tuple of Torrent objects with requested torrents
            success:  False if no torrents were found, True otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        if torrents is None:
            return await self._get_torrents_by_ids(keys)
        elif isinstance(torrents, (str, TorrentFilter)):
            return await self._get_torrents_by_filter(keys, tfilter=torrents)
        elif isinstance(torrents, abc.Sequence) and \
             all(isinstance(id, int) for id in torrents):
            return await self._get_torrents_by_ids(keys, ids=torrents)
        else:
            raise ValueError("Invalid 'torrents' argument: %r" % (torrents,))


    async def _torrent_action(self, method, torrents=None, method_args={},
                              check=None, check_keys=()):
        """
        Helper method that operates on torrents (start, stop, remove, etc)

        method:      Any method from TransmissionRPC that accepts torrent ids
        torrents:    See `torrents` method
        method_args: Dictionary with keyword arguments for method (except 'ids')
        check:       None or callable that is called with every torrent; must
                     return a 2-tuple of (SUCCESS, MESSAGE) where SUCCESS is
                     evaluated as bool and MESSAGE a string or None.  If SUCCESS
                     evaluates to True, `method` is applied to the torrent,
                     otherwise not.
        check_keys:  List of Torrent keys the check function needs ('id' and
                     'name' are always included)

        Return Response with the following properties:
            torrents: Tuple of Torrents that `method` was applied to with the
                      keys 'id' and 'name'
            success:  True if `method` was successfully applied to at least one
                      torrent, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        # Always provide some basic keys
        check_keys = set(tuple(check_keys) + ('id', 'name'))

        msgs = []
        errors = []
        response = await self.torrents(torrents, keys=check_keys)
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            msgs.extend(response.msgs)
            errors.extend(response.errors)

        if check is None:
            tlist = response.torrents
        else:
            # Filter torrents through check function
            tlist = []
            for t in response.torrents:
                passed, msg = check(t)
                if passed:
                    tlist.append(t)
                    if msg is not None:
                        msgs.append(msg)
                else:
                    if msg is not None:
                        errors.append(str(msg))

        # Apply method to torrents that passed the check function
        if len(tlist) <= 0:
            return Response(success=False, torrents=(), msgs=msgs, errors=errors)
        else:
            try:
                # Ignore response because it is always {}, except for
                # 'torrent-get' requests, which this method is not meant for.
                ids = tuple(t['id'] for t in tlist)
                # log.debug('Sending %s(%s) for IDs: %s', method.__qualname__,
                #           ', '.join(('%s=%r' % (k,v) for k,v in method_args.items())),
                #           ids)
                await method(ids=ids, **method_args)
            except ClientError as e:
                errors.append(str(e))
                return Response(success=False, torrents=(), msgs=msgs, errors=errors)
            else:
                return Response(success=True, torrents=tuple(tlist), msgs=msgs, errors=errors)

    async def stop(self, torrents):
        """
        Stop down-/uploading torrents

        torrents: See `torrents` method

        Return Response with the following properties:
            torrents: Tuple of stopped Torrents with the keys 'id' and 'name'
            success:  True if any torrents were stopped, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        def check(t):
            if t['status'].STOPPED in t['status']:
                return (False, 'Already stopped: ' + t['name'])
            else:
                return (True, 'Stopping ' + t['name'])

        return await self._torrent_action(self.rpc.torrent_stop, torrents,
                                          check=check, check_keys=('status',))

    async def start(self, torrents, force=False):
        """
        Start down-/uploading torrents

        torrents: See `torrents` method
        force:    Start downloading even if download queue is active and full

        Return Response with the following properties:
            torrents: Tuple of started Torrents with the keys 'id' and 'name'
            success:  True if any torrents were started, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        def check(t):
            if t['status'].STOPPED in t['status']:
                return (True, 'Starting ' + t['name'])
            else:
                return (False, 'Already started: ' + t['name'])

        if force:
            method = self.rpc.torrent_start_now
        else:
            method = self.rpc.torrent_start

        return await self._torrent_action(method, torrents,
                                          check=check, check_keys=('status',),
                                          method_args={'force':force})

    async def toggle_stopped(self, torrents, force=False):
        """
        Toggle down-/uploading torrents

        torrents: See `torrents` method
        force:    See `start` method

        Return Response with the following properties:
            torrents: Tuple of toggled Torrents with the keys 'id' and 'name'
            success:  True if any torrents were toggled, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        response = await self.torrents(torrents, keys=('status',))
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)

        stopped, running = [], []
        for t in response.torrents:
            if t['status'].STOPPED in t['status']:
                stopped.append(t)
            else:
                running.append(t)

        torrents = []
        msgs = []
        errors = []
        if len(running) > 0:
            response = await self.stop(tuple(t['id'] for t in running))
            torrents.extend(response.torrents)
            msgs.extend(response.msgs)
            errors.extend(response.errors)
        if len(stopped) > 0:
            r = await self.start(tuple(t['id'] for t in stopped), force=force)
            torrents.extend(response.torrents)
            msgs.extend(response.msgs)
            errors.extend(response.errors)

        return Response(success=len(torrents) > 0, torrents=torrents,
                        msgs=msgs, errors=errors)

    async def verify(self, torrents):
        """
        Verify torrents' downloaded data

        torrents: See `torrents` method

        Return Response with the following properties:
            torrents: tuple of to be verified Torrents with the keys 'id' and 'name'
            success:  True if any torrents are going to be verified, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        def check(t):
            if t['status'].VERIFY in t['status']:
                if t['status'].QUEUED in t['status']:
                    return (False, 'Already queued for verification: ' + t['name'])
                else:
                    return (False, 'Already verifying: ' + t['name'])
            else:
                return (True, 'Verifying ' + t['name'])

        return await self._torrent_action(self.rpc.torrent_verify, torrents,
                                          check=check, check_keys=('status',))

    async def remove(self, torrents, delete=False):
        """
        Remove torrents

        torrents: See `torrents` method
        delete:   True to deleted downloaded files

        Return Response with the following properties:
            torrents: Tuple of removed Torrents with the keys 'id' and 'name'
            success:  True if any torrents were removed, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        if delete:
            msg = 'Deleting %s (including files)'
        else:
            msg = 'Removing %s (keeping files)'

        def create_info_msg(t):
            return (True, msg % t['name'])

        return await self._torrent_action(self.rpc.torrent_remove, torrents,
                                          check=create_info_msg,
                                          method_args={'delete-local-data': delete})


    async def move(self, torrents, destination):
        """
        Change torrents' file system location

        torrents:    See `torrents` method
        destination: New path (relative paths are relative to the default
                     download path

        Return Response with the following properties:
            torrents: Tuple of moved Torrents with the keys 'id' and 'name'
            success:  True if any torrents were moved, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        # Transmission wants an absolute path
        response = await self._abs_download_path(destination)
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            destination = response.path

        def create_info_msg(t):
            if t['path'] != destination:
                return (True, 'Moving to %s: %s' % (destination, t['name']))
            else:
                return (False, 'Already in %s: %s' % (destination, t['name']))

        return await self._torrent_action(self.rpc.torrent_set_location, torrents,
                                          check=create_info_msg, check_keys=('path',),
                                          method_args={'move': True, 'location': destination})

    async def rename(self, tid, path, new_name):
        """
        Change the name of a torrent or one of its files or directories

        tid:      Torrent ID
        path:     Relative path in the torrent without the torrent's name or
                  anything that evaluates to False to rename the torrent itself
        new_name: New file or directory name; must not contain any directory
                  separators (usually "/") or be "." or ".."

        Return Response with the following properties:
            torrent: Torrent object with the keys 'id', 'name' and 'files' or
                     None in case of failure
            success: True if the torrent was renamed, False otherwise
            msgs:    List of info messages
            errors:  List of error messages
        """
        # Validate new_name
        if '/' in new_name:
            return Response(success=False, torrent=None,
                            errors=('New name must not contain "/": %s' % new_name,))
        for forbidden in ('.', '..'):
            if new_name == forbidden:
                return Response(success=False, torrent=None,
                                errors=('Illegal name: %s' % new_name,))

        # Fetch torrent
        response = await self._get_torrents_by_ids(ids=(tid,),
                                                   keys=('name', 'id', 'files', 'path'))
        if not response.success:
            return Response(success=False, torrent=None, errors=response.errors)
        else:
            torrent = response.torrents[0]

        # If no path given, rename the torrent itself
        if not path:
            path = torrent['name']

            # Check if new_name already exists at torrent's path
            download_path = torrent['path']
            response = await self.torrents('path=%s' % download_path, keys=('name',))
            if not response.success:
                return Response(success=False, torrent=None, errors=response.errors)
            else:
                for t in response.torrents:
                    if t['name'] == new_name:
                        return Response(success=False, torrent=None,
                                        errors=('Torrent already exists in %s: %s' %
                                                (download_path, new_name),))

        # Rename a file or directory in the torrent
        else:
            # Prepend torrent name to path
            path = os.path.join(torrent['name'], os.path.normpath(path))

            # Create list of existing files and directories; make sure to
            # include directories that contain no files (that means we can't use
            # os.path.dirname(filepath))
            existing_paths = set()
            for file in torrent['files'].files:
                filepath = str(file['path-relative'])
                existing_paths.add(filepath)
                dirs = filepath.split(os.sep)
                for i in range(1, len(dirs)):
                    dirpath = os.path.join(*dirs[:i])
                    existing_paths.add(dirpath)

            # Check if old path exists in torrent's files
            if path not in existing_paths:
                return Response(success=False, torrent=None,
                                errors=('No such path: %s' % path,))

            # Check if path is already named new_name
            if os.path.basename(path) == new_name:
                return Response(success=False, torrent=None,
                                errors=('Already named %s: %s' % (new_name, path),))

            # Check if new_name would overwrite a file or directory
            new_filepath = os.path.join(os.path.dirname(path), new_name)
            if new_filepath in existing_paths:
                return Response(success=False, torrent=None,
                                errors=('Path already exists: %s' % new_filepath,))

        # Send the rename RPC call
        def create_info_msg(t):
            return (True, 'Renaming %s to %s' % (path, new_name))
        response = await self._torrent_action(self.rpc.torrent_rename_path, (torrent['id'],),
                                              check=create_info_msg, check_keys=('name',),
                                              method_args={'path': path, 'name': new_name})
        if not response.success:
            return Response(success=False, torrent=None, errors=response.errors)
        else:
            # Preserve info messages for final response
            msgs = response.msgs

        # Fetch new torrent data and return final response
        response = await self._get_torrents_by_ids(ids=(tid,),
                                                   keys=('name', 'id', 'files'))
        if not response.success:
            return Response(success=False, torrent=None, errors=response.errors)
        else:
            torrent = response.torrents[0]
            return Response(success=True, torrent=torrent, msgs=msgs)

    async def file_priority(self, torrents, files, priority):
        """
        Change download priority of individual torrent files

        torrents: See `torrents` method
        files:    TorrentFileFilter object (or its string representation), sequence
                  of (torrent ID, file ID) tuples or None for all files
        priority: One of the strings 'off', 'low', 'normal' or 'high'

        Return Response with the following properties:
            torrents: Tuple of matching Torrents with matching files with the
                      keys 'id', 'name' and 'files'
            success:  True if any file priorities were changed, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        response = await self.torrents(torrents, keys=('name', 'files'))
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            if isinstance(files, str):
                files = TorrentFileFilter(files)

            # Set filter_files to a lambda that takes a TorrentFileTree and
            # returns a list of TorrentFiles.
            if files is None:
                filter_files = lambda ftree: tuple(ftree.files)
            elif isinstance(files, TorrentFileFilter):
                filter_files = lambda ftree: tuple(files.apply(ftree.files))
            elif isinstance(files, abc.Sequence):
                filter_files = lambda ftree: tuple(f for f in ftree.files
                                                   if f['id'] in files)
            else:
                raise ValueError("Invalid 'files' argument: %r" % (files,))

            torrent_ids = []
            msgs = []
            errors = []
            for t in sorted(response.torrents, key=lambda t: t['name'].lower()):
                # Filter torrent's files
                flist = filter_files(t['files'])
                if files is None:
                    msgs.append('%d file%s: %s' %
                                (len(flist), '' if len(flist) == 1 else 's', t['name']))
                else:
                    if not flist:
                        errors.append('No matching files: %s' % (t['name'],))
                    else:
                        msgs.append('%d matching file%s: %s' %
                                    (len(flist), '' if len(flist) == 1 else 's', t['name']))
                success = len(flist) > 0

                # Transmission wants a list of file indexes.  For
                # aiotransmission, the 'id' field of a TorrentFile is a tuple:
                #     (<torrent ID>, <file index>)
                # (See aiotransmission.torrent._create_TorrentFileTree())
                findexes = tuple(f['id'][1] for f in flist)
                if findexes:
                    response = await self._set_files_priority(priority, t['id'], findexes)
                    if response.success:
                        torrent_ids.append(t['id'])
                    msgs.extend(response.msgs)
                    errors.extend(response.errors)

        if torrent_ids:
            response = await self.torrents(torrent_ids, keys=('id', 'name', 'files'))
            if not response.success:
                return Response(success=False, torrents=(), errors=response.errors)
            else:
                torrents = response.torrents
        return Response(success=success, torrents=torrents, msgs=msgs, errors=errors)

    async def _set_files_priority(self, priority, torrent_id, file_indexes):
        fi = tuple(file_indexes)
        log.debug('Setting priority of torrent #%d: %r: %s', torrent_id, priority, file_indexes)
        if priority in ('high', 'normal', 'low'):
            return await self._torrent_action(
                self.rpc.torrent_set, (torrent_id,),
                method_args={'priority-%s' % priority: fi, 'files-wanted': fi})
        elif priority == 'off':
            return await self._torrent_action(
                self.rpc.torrent_set, (torrent_id,),
                method_args={'files-unwanted': fi})
        else:
            raise ValueError('Invalid priority: {!r}'.format(priority))


    async def _limit_rate_absolute(self, torrents, direction, limit):
        if isinstance(limit, str):
            try:
                limit = BoolOrBandwidth(limit)
            except ValueError as e:
                return Response(success=False, torrents=(), errors=('%s: %r' % (e, limit),))
        if isinstance(limit, (float, int)):
            limit = BoolOrBandwidth(False) if limit < 0 or limit >= float('inf') else limit

        log.debug('Setting new %sload limit for torrents %s: %s', direction, torrents, limit)
        return await self._limit_rate(torrents, direction, get_new_limit=lambda _: limit)

    async def _limit_rate_relative(self, torrents, direction, adjustment):
        if isinstance(adjustment, str):
            try:
                adjustment = Bandwidth(adjustment)
            except ValueError as e:
                return Response(success=False, torrents=(), errors=('%s: %r' % (e, adjustment),))

        def add_to_current_limit(current_limit):
            log.debug('Adjusting %sload limit %r by %r', direction, current_limit, adjustment)
            return BoolOrBandwidth.adjust(current_limit, adjustment)

        return await self._limit_rate(torrents, direction, get_new_limit=add_to_current_limit)

    async def _limit_rate(self, torrents, direction, get_new_limit):
        response = await self._map_tid_to_torrent_values(torrents, keys=('limit-rate-'+direction,))
        if not response.success:
            return Response(success=False, torrent_set_args={}, errors=response.errors)
        else:
            current_limits = response.torrent_values
            log.debug('Current %sload rate limits: %r', direction, current_limits)

        # Generate 'torrent-set' arguments for each torrent ID.  To de-duplicate
        # requests (same args for multiple torrents), we map the args to a list
        # of torrent IDs, so we just append another ID for existing args.
        torrent_set_args = {}
        errors = {}
        for tid,cur_limit in current_limits.items():
            cur_limit = BoolOrBandwidth(cur_limit)
            new_limit = BoolOrBandwidth(get_new_limit(cur_limit))

            if new_limit == cur_limit:
                log.debug('Nothing to set: new:%r == cur:%r', new_limit, cur_limit)
                errors[tid] = 'Already %s' % cur_limit
                continue
            elif isinstance(new_limit, Bool):
                if new_limit and isinstance(cur_limit, (float, int)) and cur_limit < float('inf'):
                    errors[tid] = 'Already limited'
                    continue
                else:
                    log.debug('%sabling limit', 'En' if new_limit else 'Dis')
                    args = (('%sloadLimited' % direction, bool(new_limit)),)
            else:
                log.debug('Setting new limit: %r', new_limit)
                if new_limit >= float('inf'):
                    args = (('%sloadLimited' % direction, False),)
                else:
                    raw_limit = round(int(new_limit.copy(convert_to='B'))/1000)  # Transmission expects kilobytes
                    args = (('%sloadLimited' % direction, True),
                            ('%sloadLimit' % direction, raw_limit))

            if args in torrent_set_args:
                torrent_set_args[args].append(tid)
            else:
                torrent_set_args[args] = [tid]

        # Send one 'torrent-set' request for each list of torrent IDs
        for args,tids in torrent_set_args.items():
            response = await self._torrent_action(self.rpc.torrent_set, tids,
                                                  method_args=dict(args))
            if not response.success:
                return Response(success=False, torrents=(), errors=response.errors)

        # Fetch torrents again and return Response with new rate limit messages
        all_tids = sum(torrent_set_args.values(), []) + list(errors)
        response = await self.torrents(all_tids, keys=('name', 'id', 'limit-rate-'+direction))
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        msgs = []
        errormsgs = []
        success = False
        for t in response.torrents:
            if t['id'] in errors:
                errormsgs.append('%s %sload rate limit: %s' % (t['name'], direction, errors[t['id']]))
            else:
                success = True
                msgs.append('%s %sload rate limit: %s' % (t['name'], direction, t['limit-rate-'+direction]))
        return Response(success=success, torrents=response.torrents, msgs=msgs, errors=errormsgs)

    async def set_limit_rate_up(self, torrents, limit):
        """
        Limit upload rate for individual torrent(s)

        torrents: See `torrents` method
        limit:    Passed to `client.utils.BoolOrBandwidth`

        Return Response with the following properties:
            torrents: Tuple of Torrents with the keys 'id', 'name' and 'limit-rate-up'
            success:  True if any upload rate limits were changed, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        return await self._limit_rate_absolute(torrents, 'up', limit)

    async def set_limit_rate_down(self, torrents, limit):
        """See `set_limit_rate_up`"""
        return await self._limit_rate_absolute(torrents, 'down', limit)

    async def adjust_limit_rate_up(self, torrents, adjustment):
        """
        Same as `set_limit_rate_up` but set new limit relative to current limit

        adjustment: Negative or positive number to add to the current limit of
                    each matching torrent (this is passed to
                    `client.utils.Bandwidth`)
        """
        return await self._limit_rate_relative(torrents, 'up', adjustment)

    async def adjust_limit_rate_down(self, torrents, adjustment):
        """See `adjust_limit_rate_up`"""
        return await self._limit_rate_relative(torrents, 'down', adjustment)


    async def tracker_add(self, torrents, urls):
        """
        Add tracker(s) to torrents

        torrents: See `torrents` method
        urls:     Iterable of announce URLs

        Return Response with the following properties:
            torrents: Tuple of Torrents with the keys 'id', 'name' and 'trackers'
            success:  True if any trackers were added, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        if not urls:
            return Response(success=False, torrents=(), errors=('No URLs given',))

        # Transmission returns 'Invalid argument' if we try to add an existing
        # tracker, so first we check if any of our URLs already exist.
        response = await self.torrents(torrents, keys=('id', 'name', 'trackers',))
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            tordict = {tor['id']:tor for tor in response.torrents}

        # Map torrent IDs to currently used URLs by that torrent
        old_url_dict = {torid:tuple(trk['url-announce'] for trk in torrent['trackers'])
                        for torid,torrent in tordict.items()}

        # Make sure URLs are comparable
        new_urls = [URL(url) for url in urls]

        # For each torrent, report any supplied URLs that are already used
        msgs = []
        errors = []
        for torid,old_urls in old_url_dict.items():
            for old_url in old_urls:
                if old_url in new_urls:
                    errors.append('%s: Tracker already exists: %s' % (tordict[torid]['name'], old_url))
                    new_urls.remove(old_url)

        # No URLs left to add?
        if not new_urls:
            return Response(success=False, torrents=(), errors=errors)

        for new_url in new_urls:
            msgs.append('%s: Adding tracker: %s' % (tordict[torid]['name'], new_url))

        # Add trackers
        args = {'trackerAdd': [str(url) for url in new_urls]}
        response = await self._torrent_action(self.rpc.torrent_set, torrents,
                                              method_args=args)
        if not response.success:
            errors.extend(response.errors)
            return Response(success=False, torrents=(), msgs=msgs, errors=errors)
        else:
            return Response(success=True, torrents=response.torrents, msgs=msgs, errors=errors)

    async def tracker_remove(self, torrents, urls, partial_match=False):
        """
        Remove tracker(s) from torrents

        torrents:      See `torrents` method
        urls:          Iterable of announce URLs
        partial_match: True if given URLs match existing URLs partially
                       (e.g. 'example.org' matches 'http://tracker.example.org/')

        Return Response with the following properties:
            torrents: Tuple of Torrents with the keys 'id', 'name' and 'trackers'
            success:  True if any trackers were removed, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        if not urls:
            return Response(success=False, torrents=(), errors=('No URLs given',))

        # Get wanted torrent IDs
        response = await self.torrents(torrents, keys=('id',))
        if not response.success:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            torids = tuple(t['id'] for t in response.torrents)

        # Get raw tracker lists for the unaltered tracker IDs.  We need them
        # later to specify which trackers to remove.
        response = await self._request(self.rpc.torrent_get, ids=torids,
                                       fields=('id', 'name', 'trackers'))
        if not response.success or len(response.result) <= 0:
            return Response(success=False, torrents=(), errors=response.errors)
        else:
            raw_tor_dict = {raw_tor['id']:raw_tor for raw_tor in response.result}

        # Map torrent IDs to lists of IDs of matching trackers
        msgs = []
        errors = []
        remove_urls = urls
        matching_urls = []
        remove_ids = {}
        for torid,raw_tor in raw_tor_dict.items():
            remove_ids[torid] = []
            for raw_trk in raw_tor['trackers']:
                existing_url = raw_trk['announce']
                for remove_url in remove_urls:
                    if remove_url == existing_url or partial_match and remove_url in existing_url:
                        remove_ids[torid].append(raw_trk['id'])
                        matching_urls.append(remove_url)
                        msgs.append('%s: Removing tracker: %s' % (raw_tor['name'], existing_url))

            if len(remove_ids[torid]) <= 0:
                # No matching trackers for this torrent
                del remove_ids[torid]

        # Report error if no matching trackers were found for a given URL
        for mismatch in set(remove_urls).difference(matching_urls):
            errors.append('No matching trackers found: %r' % mismatch)

        # Finally remove trackers from torrents
        if remove_ids:
            for torid,trkids in remove_ids.items():
                response = await self._torrent_action(self.rpc.torrent_set, (torid,),
                                                      method_args={'trackerRemove': trkids})
                if not response.success:
                    return Response(success=False, torrents=(), errors=response.errors)

        # Get new torrent list with newly added trackers
        response = await self.torrents(tuple(remove_ids), keys=('id', 'name', 'trackers'))
        if not response.success:
            errors.extend(response.errors)
            return Response(success=False, torrents=(), msgs=msgs, errors=errors)
        else:
            return Response(success=True, torrents=response.torrents, msgs=msgs, errors=errors)

    async def announce(self, torrents):
        """
        Announce torrents to their tracker(s)

        torrents: See `torrents` method

        Return Response with the following properties:
            torrents: Tuple of Torrents with the keys 'id' and 'name'
            success:  True if any torrents were found, False otherwise
            msgs:     List of info messages
            errors:   List of error messages
        """
        import time
        def check(t):
            if len(t['trackers']) < 1:
                return (False, 'Torrent has no trackers: %s' % t['name'])
            elif t['status'].STOPPED in t['status']:
                return (False, 'Not announcing inactive torrent: %s' % t['name'])
            elif t['time-manual-announce-allowed'] == 0:
                return (False, 'Manual announce not allowed')
            elif t['time-manual-announce-allowed'] > time.time():
                return (False, ('Manual announce not allowed until %s (in %s): %r' %
                                (t['time-manual-announce-allowed'],
                                 t['time-manual-announce-allowed'].timedelta, t['name'])))
            else:
                return (True, 'Announcing: %s' % t['name'])

        return await self._torrent_action(self.rpc.torrent_reannounce, torrents,
                                          check=check, check_keys=('status', 'trackers',
                                                                   'time-manual-announce-allowed'))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import os
from xdg.BaseDirectory import xdg_config_home as XDG_CONFIG_HOME
from xdg.BaseDirectory import xdg_cache_home  as XDG_CACHE_HOME
from xdg.BaseDirectory import xdg_data_home  as XDG_DATA_HOME

from .. import __appname__
from ..views import (torrent, file, peer, tracker, setting)
from ..client.sorters.torrent import TorrentSorter
from ..client.sorters.peer import TorrentPeerSorter
from ..client.sorters.tracker import TorrentTrackerSorter
from ..client.sorters.setting import SettingSorter

DEFAULT_RCFILE      = os.path.join(XDG_CONFIG_HOME, __appname__, 'rc')
DEFAULT_HISTORY_DIR = os.path.join(XDG_DATA_HOME, __appname__, 'histories')
DEFAULT_GEOIP_DIR   = os.path.join(XDG_CACHE_HOME, __appname__)
DEFAULT_THEME_FILE  = os.path.join(os.path.dirname(__file__), 'default.theme')


def init_defaults(localcfg):
    from ..utils.usertypes import (String, Int, Float, Bool, Path, Tuple, Option)

    class SortOrder(str):
        """String that is equal to the same string with '!' or '.' prepended"""
        _invert_chars = ''.join(TorrentSorter.INVERT_CHARS)
        def __eq__(self, other):
            return self.lstrip(self._invert_chars) == other.lstrip(self._invert_chars)

        # An overloaded __eq__() obligates an overloaded __hash__(), or
        # instances won't be hashable.
        def __hash__(self):
            return super().__hash__()


    def partial_sort_order(sortercls):
        options = tuple(SortOrder(opt) for opt in sortercls.SORTSPECS)
        return Tuple.partial(options=options, dedup=True)

    localcfg.add('connect.host',
                 constructor=String.partial(),
                 default='localhost',
                 description='Hostname or IP of Transmission RPC interface')
    localcfg.add('connect.port',
                 Int.partial(min=1, max=65535, prefix='none'),
                 default=9091,
                 description='Port of Transmission RPC interface')
    localcfg.add('connect.path',
                 String.partial(),
                 default='/transmission/rpc',
                 description='Path of Transmission RPC interface')
    localcfg.add('connect.user',
                 String.partial(),
                 default='',
                 description='Username to use for authentication with Transmission RPC interface')
    localcfg.add('connect.password',
                 String.partial(),
                 default='',
                 description='Password to use for authentication with Transmission RPC interface')
    localcfg.add('connect.timeout',
                 Float.partial(min=0),
                 default=10,
                 description='Number of seconds before connecting to Transmission RPC interface fails')
    localcfg.add('connect.tls',
                 Bool.partial(),
                 default='off',
                 description='Whether to connect via HTTPS to the Transmission RPC interface')

    localcfg.add('columns.torrents',
                 Tuple.partial(options=torrent.COLUMNS, aliases=torrent.ALIASES),
                 default=('marked', 'size', 'downloaded', 'uploaded', 'ratio',
                          'seeds', 'peers', 'status', 'eta', '%downloaded',
                          'rate-down', 'rate-up', 'name'),
                 description='Default columns in torrent lists')
    localcfg.add('columns.peers',
                 Tuple.partial(options=peer.COLUMNS, aliases=peer.ALIASES),
                 default=('host', 'client', 'country', '%downloaded', 'rate-down',
                          'rate-up', 'rate-est', 'eta'),
                 description='Default columns in peer lists')
    localcfg.add('columns.files',
                 Tuple.partial(options=file.COLUMNS, aliases=file.ALIASES),
                 default=('marked', 'priority', '%downloaded', 'downloaded', 'size', 'name'),
                 description='Default columns in file lists')
    localcfg.add('columns.trackers',
                 Tuple.partial(options=tracker.COLUMNS, aliases=tracker.ALIASES),
                 default=('tier', 'domain', 'error', 'last-announce', 'next-announce',
                          'leeches', 'seeds', 'downloads'),
                 description='Default columns in tracker lists')
    localcfg.add('columns.settings',
                 Tuple.partial(options=setting.COLUMNS, aliases=setting.ALIASES),
                 default=('name', 'value', 'default', 'description'),
                 description='Default columns in setting lists')

    localcfg.add('geoip',
                 Bool.partial(),
                 default=True,
                 description='Whether to lookup peers\' country codes')
    localcfg.add('geoip.dir',
                 Path.partial(),
                 default=DEFAULT_GEOIP_DIR,
                 description='Where to cache the downloaded geolocation database')

    localcfg.add('remove.max-hits',
                 Int.partial(),
                 default=10,
                 description=('Maximum number of torrents to remove without extra confirmation'
                              '; negative numbers mean "unlimited"'))

    localcfg.add('reverse-dns',
                 Bool.partial(),
                 default=True,
                 description=('Whether to lookup peers\' host names'))

    localcfg.add('sort.torrents',
                 partial_sort_order(TorrentSorter),
                 default=TorrentSorter.DEFAULT_SORT,
                 description='List of sort orders in torrent lists')
    localcfg.add('sort.peers',
                 partial_sort_order(TorrentPeerSorter),
                 default=TorrentPeerSorter.DEFAULT_SORT,
                 description='List of sort orders in peer lists')
    localcfg.add('sort.trackers',
                 partial_sort_order(TorrentTrackerSorter),
                 default=TorrentTrackerSorter.DEFAULT_SORT,
                 description='List of sort orders in tracker lists')
    localcfg.add('sort.settings',
                 partial_sort_order(SettingSorter),
                 default=SettingSorter.DEFAULT_SORT,
                 description='List of sort orders in setting lists')

    localcfg.add('tui.theme',
                 Path.partial(),
                 default=DEFAULT_THEME_FILE,
                 description='Path to theme file'),
    localcfg.add('tui.log.height',
                 Int.partial(min=1),
                 default=10,
                 description='Maximum height of the log section')
    localcfg.add('tui.log.autohide',
                 Float.partial(min=0),
                 default=10,
                 description=('If the log is hidden, show it for this many seconds '
                              'for new log entries before hiding it again'))
    localcfg.add('tui.cli.history-dir',
                 Path.partial(),
                 default=DEFAULT_HISTORY_DIR,
                 description='Directory where histories of user input are stored')
    localcfg.add('tui.cli.history-size',
                 Int.partial(min=0),
                 default=10000,
                 description='Maximum number of lines to keep in history files')
    localcfg.add('tui.poll',
                 Float.partial(min=0.1),
                 default=5,
                 description='Interval in seconds between TUI updates')

    localcfg.add('unit.bandwidth',
                 Option.partial(options=('bit', 'byte')),
                 default='byte',
                 description="Unit for bandwidth rates ('bit' or 'byte')")
    localcfg.add('unitprefix.bandwidth',
                 Option.partial(options=('metric', 'binary')),
                 default='metric',
                 description=("Unit prefix for bandwidth rates ('metric' or 'binary')"))

    localcfg.add('unit.size',
                 Option.partial(options=('bit', 'byte')),
                 default='byte',
                 description="Unit for file sizes ('bit' or 'byte')")
    localcfg.add('unitprefix.size',
                 Option.partial(options=('metric', 'binary')),
                 default='binary',
                 description=("Unit prefix for file sizes ('metric' or 'binary')"))

    localcfg.add('tui.marked.on',
                 String.partial(minlen=1, maxlen=1),
                 default='●',
                 description=("Character displayed in 'marked' column for marked "
                              "list items (see 'mark' command)"))
    localcfg.add('tui.marked.off',
                 String.partial(minlen=1, maxlen=1),
                 default=' ',
                 description=("Character displayed in 'marked' column for unmarked "
                              "list items (see 'mark' command)"))



DEFAULT_KEYMAP = (
    # Some vi and emacs key translations
    {'key': 'h',      'action': '<left>'},
    {'key': 'j',      'action': '<down>'},
    {'key': 'k',      'action': '<up>'},
    {'key': 'l',      'action': '<right>'},
    {'key': 'g',      'action': '<home>'},
    {'key': 'G',      'action': '<end>'},
    {'key': 'ctrl-n', 'action': '<down>'},
    {'key': 'ctrl-p', 'action': '<up>'},
    {'key': 'ctrl-f', 'action': '<right>'},
    {'key': 'ctrl-b', 'action': '<left>'},

    # Global TUI keys
    {'context': 'main', 'key': 'q',     'action': 'quit',
     'description': 'Exit'},
    {'context': 'main', 'key': ':',     'action': 'tui show cli',
     'description': 'Open command line interface'},
    {'context': 'main', 'key': 'alt-s', 'action': 'tab set',
     'description': 'Open settings in a new tab'},

    # Help
    {'context': 'main', 'key': 'F1+c', 'action': 'tab help commands',
     'description': 'Open help for commands in a new tab'},
    {'context': 'main', 'key': 'F1+s', 'action': 'tab help settings',
     'description': 'Open help for settings in a new tab'},
    {'context': 'main', 'key': 'F1+k', 'action': 'tab help keymap',
     'description': 'Open help for keybindings in a new tab'},
    {'context': 'main', 'key': 'F1+f', 'action': 'tab help filters',
     'description': 'Open help for filters in a new tab'},
    {'context': 'main', 'key': 'F1+r', 'action': 'tab help rcfile',
     'description': 'Open help for rc files in a new tab'},
    {'context': 'main', 'key': '?',    'action': '<F1>'},

    # Hide/Show TUI elements
    {'context': 'main', 'key': 'alt-m', 'action': 'tui toggle main',
     'description': 'Show or hide main area'},
    {'context': 'main', 'key': 'alt-t', 'action': 'tui toggle topbar',
     'description': 'Show or hide top bar'},
    {'context': 'main', 'key': 'alt-b', 'action': 'tui toggle bottombar',
     'description': 'Show or hide bottom bar'},
    {'context': 'main', 'key': 'alt-l', 'action': 'tui toggle log',
     'description': 'Show or hide log messages'},

    # Log messages
    {'context': 'main', 'key': 'ctrl-l',   'action': 'log clear',
     'description': 'Remove all log messages'},
    {'context': 'main', 'key': 'alt-pgup', 'action': 'log scroll page up',
     'description': 'Scroll log messages one page up'},
    {'context': 'main', 'key': 'alt-pgdn', 'action': 'log scroll page down',
     'description': 'Scroll log messages one page down'},
    {'context': 'main', 'key': 'alt-home', 'action': 'log scroll top',
     'description': 'Scroll to top of log messages'},
    {'context': 'main', 'key': 'alt-end',  'action': 'log scroll bottom',
     'description': 'Scroll to bottom of log messages'},

    # Global bandwidth limits
    {'context': 'main', 'key': 'shift-up',    'action': 'rate --quiet dn -- +=100kB global',
     'description': 'Increase global download rate limit by 100 kilobytes'},
    {'context': 'main', 'key': 'shift-down',  'action': 'rate --quiet dn -- -=100kB global',
     'description': 'Decrease global download rate limit by 100 kilobytes'},
    {'context': 'main', 'key': 'shift-right', 'action': 'rate --quiet up -- +=100kB global',
     'description': 'Increase global upload rate limit by 100 kilobytes'},
    {'context': 'main', 'key': 'shift-left',  'action': 'rate --quiet up -- -=100kB global',
     'description': 'Decrease global upload rate limit by 100 kilobytes'},

    # Tab management
    {'context': 'tabs', 'key': 'n',     'action': 'tab',
     'description': 'Open a new empty tab'},
    {'context': 'tabs', 'key': 'd',     'action': 'tab --close',
     'description': 'Remove current tab and focus tab on the right'},
    {'context': 'tabs', 'key': 'D',     'action': 'tab --close --focus left',
     'description': 'Remove current tab and focus tab on the left'},
    {'context': 'tabs', 'key': 'alt-1', 'action': 'tab --focus 1',
     'description': 'Focus first tab'},
    {'context': 'tabs', 'key': 'alt-2', 'action': 'tab --focus 2',
     'description': 'Focus second tab'},
    {'context': 'tabs', 'key': 'alt-3', 'action': 'tab --focus 3',
     'description': 'Focus third tab'},
    {'context': 'tabs', 'key': 'alt-4', 'action': 'tab --focus 4',
     'description': 'Focus fourth tab'},
    {'context': 'tabs', 'key': 'alt-5', 'action': 'tab --focus 5',
     'description': 'Focus fifth tab'},
    {'context': 'tabs', 'key': 'alt-6', 'action': 'tab --focus 6',
     'description': 'Focus sixth tab'},
    {'context': 'tabs', 'key': 'alt-7', 'action': 'tab --focus 7',
     'description': 'Focus seventh tab'},
    {'context': 'tabs', 'key': 'alt-8', 'action': 'tab --focus 8',
     'description': 'Focus eighth tab'},
    {'context': 'tabs', 'key': 'alt-9', 'action': 'tab --focus 9',
     'description': 'Focus ninth tab'},
    {'context': 'tabs', 'key': 'alt-0', 'action': 'tab --focus 10',
     'description': 'Focus tenth tab'},

    # List torrents with different filters
    {'context': 'tabs', 'key': 'f a', 'action': 'tab ls active',
     'description': 'List active torrents in a new tab'},
    {'context': 'tabs', 'key': 'f A', 'action': 'tab ls !active',
     'description': 'List inactive torrents in a new tab'},
    {'context': 'tabs', 'key': 'f i', 'action': 'tab ls isolated --columns name,tracker,error',
     'description': 'List isolated torrents in a new tab'},
    {'context': 'tabs', 'key': 'f p', 'action': 'tab ls paused',
     'description': 'List paused torrents in a new tab'},
    {'context': 'tabs', 'key': 'f P', 'action': 'tab ls !paused',
     'description': 'List unpaused torrents in a new tab'},
    {'context': 'tabs', 'key': 'f c', 'action': 'tab ls complete',
     'description': 'List complete torrents in a new tab'},
    {'context': 'tabs', 'key': 'f C', 'action': 'tab ls !complete',
     'description': 'List incomplete torrents in a new tab'},
    {'context': 'tabs', 'key': 'f u', 'action': 'tab ls uploading',
     'description': 'List uploading torrents in a new tab'},
    {'context': 'tabs', 'key': 'f d', 'action': 'tab ls downloading',
     'description': 'List downloading torrents in a new tab'},
    {'context': 'tabs', 'key': 'f s', 'action': 'tab ls seeding',
     'description': 'List seeding torrents in a new tab'},
    {'context': 'tabs', 'key': 'f l', 'action': 'tab ls leeching',
     'description': 'List leeching torrents in a new tab'},
    {'context': 'tabs', 'key': 'f .', 'action': 'tab ls',
     'description': 'List all torrents in a new tab'},

    # Torrent list actions
    {'context': 'torrentlist', 'key': 's d', 'action': 'sort --add dir',
     'description': 'Sort torrents by download directory'},
    {'context': 'torrentlist', 'key': 's D', 'action': 'sort --add !dir',
     'description': 'Sort torrents by download directory in reverse'},
    {'context': 'torrentlist', 'key': 's e', 'action': 'sort --add eta',
     'description': 'Sort torrents by estimated remaining download time'},
    {'context': 'torrentlist', 'key': 's E', 'action': 'sort --add !eta',
     'description': 'Sort torrents by estimated remaining download time in reverse'},
    {'context': 'torrentlist', 'key': 's n', 'action': 'sort --add name',
     'description': 'Sort torrents by name'},
    {'context': 'torrentlist', 'key': 's N', 'action': 'sort --add !name',
     'description': 'Sort torrents by name in reverse'},
    {'context': 'torrentlist', 'key': 's o', 'action': 'sort --add ratio',
     'description': 'Sort torrents by upload/download ratio'},
    {'context': 'torrentlist', 'key': 's O', 'action': 'sort --add !ratio',
     'description': 'Sort torrents by upload/download ratio in reverse'},
    {'context': 'torrentlist', 'key': 's p', 'action': 'sort --add %downloaded',
     'description': 'Sort torrents by download progress'},
    {'context': 'torrentlist', 'key': 's P', 'action': 'sort --add !%downloaded',
     'description': 'Sort torrents by download progress in reverse'},
    {'context': 'torrentlist', 'key': 's r', 'action': 'sort --add rate',
     'description': 'Sort torrents by combined upload and download rate'},
    {'context': 'torrentlist', 'key': 's R', 'action': 'sort --add !rate',
     'description': 'Sort torrents by combined upload and download rate in reverse'},
    {'context': 'torrentlist', 'key': 's s', 'action': 'sort --add seeds',
     'description': 'Sort torrents by number of seeds'},
    {'context': 'torrentlist', 'key': 's S', 'action': 'sort --add !seeds',
     'description': 'Sort torrents by number of seeds in reverse'},
    {'context': 'torrentlist', 'key': 's t', 'action': 'sort --add tracker',
     'description': 'Sort torrents by first tracker domain'},
    {'context': 'torrentlist', 'key': 's T', 'action': 'sort --add !tracker',
     'description': 'Sort torrents by first tracker domain in reverse'},
    {'context': 'torrentlist', 'key': 's z', 'action': 'sort --add size',
     'description': 'Sort torrents by size of combined wanted files'},
    {'context': 'torrentlist', 'key': 's Z', 'action': 'sort --add !size',
     'description': 'Sort torrents by size of combined wanted files in reverse'},
    {'context': 'torrentlist', 'key': 's ,', 'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'torrentlist', 'key': 's .', 'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'torrentlist', 'key': '/',   'action': ("interactive 'limit \"[]\"' --per-change "
                                                        "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed torrents by applying more filters'},

    # Torrent actions
    {'context': 'torrent', 'key': 't a',       'action': 'announce',
     'description': 'Announce torrent to its tracker(s)'},
    {'context': 'torrent', 'key': 't d',       'action': 'delete',
     'description': 'Remove torrent and keep any downloaded files'},
    {'context': 'torrent', 'key': 't D',       'action': 'delete --delete-files',
     'description': 'Remove torrent and all downloaded files'},
    {'context': 'torrent', 'key': 't m',       'action': 'setcommand move {location}/',
     'description': 'Move torrent'},
    {'context': 'torrent', 'key': 't n',       'action': 'setcommand rename id={id} {name}',
     'description': 'Rename torrent'},
    {'context': 'torrent', 'key': 't p',       'action': 'tab peerlist',
     'description': 'List selected or focused torrents\' peers in a new tab'},
    {'context': 'torrent', 'key': 't t',       'action': 'tab trackerlist',
     'description': 'List selected or focused torrents\' trackers in a new tab'},
    {'context': 'torrent', 'key': 't s',       'action': 'start --toggle',
     'description': 'Start or stop selected or focused torrents'},
    # Disabled until we have download queues implemented
    # {'context': 'torrent', 'key': 't S',       'action': 'start --toggle --force',
    #  'description': ''},
    {'context': 'torrent', 'key': 't v',       'action': 'verify',
     'description': 'Verify selected or focused torrents\' downloaded files'},
    {'context': 'torrent', 'key': 'enter',     'action': 'tab summary',
     'description': 'Show details about focused torrent in a new tab'},
    {'context': 'torrent', 'key': 'alt-enter', 'action': 'tab filelist',
     'description': 'List selected or focused torrents\' files in a new tab'},
    {'context': 'torrent', 'key': 'space',     'action': 'mark --toggle --focus-next',
     'description': 'Mark or unmark focused torrent'},
    {'context': 'torrent', 'key': 'alt-space', 'action': 'mark --toggle --all',
     'description': 'Mark or unmark all torrents'},

    # Peer list actions
    {'context': 'peerlist', 'key': 's c', 'action': 'sort --add country',
     'description': 'Sort peers by country'},
    {'context': 'peerlist', 'key': 's C', 'action': 'sort --add !country',
     'description': 'Sort peers by country in reverse'},
    {'context': 'peerlist', 'key': 's d', 'action': 'sort --add rate-down',
     'description': 'Sort peers by download rate'},
    {'context': 'peerlist', 'key': 's D', 'action': 'sort --add !rate-down',
     'description': 'Sort peers by download rate in reverse'},
    {'context': 'peerlist', 'key': 's e', 'action': 'sort --add eta',
     'description': 'Sort peers by estimated time of download completion'},
    {'context': 'peerlist', 'key': 's E', 'action': 'sort --add !eta',
     'description': 'Sort peers by estimated time of download completion in reverse'},
    {'context': 'peerlist', 'key': 's l', 'action': 'sort --add client',
     'description': 'Sort peers by client'},
    {'context': 'peerlist', 'key': 's L', 'action': 'sort --add !client',
     'description': 'Sort peers by client in reverse'},
    {'context': 'peerlist', 'key': 's p', 'action': 'sort --add %downloaded',
     'description': 'Sort peers by download progress'},
    {'context': 'peerlist', 'key': 's P', 'action': 'sort --add !%downloaded',
     'description': 'Sort peers by download progress in reverse'},
    {'context': 'peerlist', 'key': 's u', 'action': 'sort --add rate-up',
     'description': 'Sort peers by upload rate'},
    {'context': 'peerlist', 'key': 's U', 'action': 'sort --add !rate-up',
     'description': 'Sort peers by upload rate in reverse'},
    {'context': 'peerlist', 'key': 's s', 'action': 'sort --add rate-est',
     'description': 'Sort peers by estimated download rate'},
    {'context': 'peerlist', 'key': 's S', 'action': 'sort --add !rate-est',
     'description': 'Sort peers by estimated download rate in reverse'},
    {'context': 'peerlist', 'key': 's r', 'action': 'sort --add rate',
     'description': 'Sort peers by combined upload and download rate'},
    {'context': 'peerlist', 'key': 's R', 'action': 'sort --add !rate',
     'description': 'Sort peers by combined upload and download rate in reverse'},
    {'context': 'peerlist', 'key': 's t', 'action': 'sort --add torrent',
     'description': 'Sort peers by torrent name'},
    {'context': 'peerlist', 'key': 's T', 'action': 'sort --add !torrent',
     'description': 'Sort peers by torrent name in reverse'},
    {'context': 'peerlist', 'key': 's ,', 'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'peerlist', 'key': 's .', 'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'peerlist', 'key': '/',   'action': ("interactive 'limit \"[]\"' --per-change "
                                                     "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed peers by applying more filters'},

    # File list actions
    {'context': 'filelist', 'key': '/', 'action': ("interactive 'limit \"[]\"' --per-change "
                                                   "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed files by applying more filters'},

    # File actions
    {'context': 'file', 'key': 'f n',       'action': 'setcommand rename {name}',
     'description': 'Rename file or directory'},
    {'context': 'file', 'key': 'f +',       'action': 'priority high',
     'description': 'Set selected or focused file\'s download priority to "high"'},
    {'context': 'file', 'key': 'f =',       'action': 'priority normal',
     'description': 'Set selected or focused file\'s download priority to "normal"'},
    {'context': 'file', 'key': 'f -',       'action': 'priority low',
     'description': 'Set selected or focused file\'s download priority to "low"'},
    {'context': 'file', 'key': 'f 0',       'action': 'priority off',
     'description': 'Don\'t download selected or focused file(s)'},
    {'context': 'file', 'key': 'space',     'action': 'mark --toggle --focus-next',
     'description': 'Mark or unmark focused file or directory'},
    {'context': 'file', 'key': 'alt-space', 'action': 'mark --toggle --all',
     'description': 'Mark or unmark all files'},

    # Tracker list actions
    {'context': 'trackerlist', 'key': 's n a', 'action': 'sort --add next-announce',
     'description': 'Sort trackers by next announce time'},
    {'context': 'trackerlist', 'key': 's n A', 'action': 'sort --add !next-announce',
     'description': 'Sort trackers by next announce time in reverse'},
    {'context': 'trackerlist', 'key': 's l a', 'action': 'sort --add last-announce',
     'description': 'Sort trackers by last announce time'},
    {'context': 'trackerlist', 'key': 's l A', 'action': 'sort --add !last-announce',
     'description': 'Sort trackers by last announce time in reverse'},
    {'context': 'trackerlist', 'key': 's d',   'action': 'sort --add downloads',
     'description': 'Sort trackers by number of downloads'},
    {'context': 'trackerlist', 'key': 's D',   'action': 'sort --add !downloads',
     'description': 'Sort trackers by number of downloads in reverse'},
    {'context': 'trackerlist', 'key': 's c',   'action': 'sort --add leeches',
     'description': 'Sort trackers by number of downloading peers'},
    {'context': 'trackerlist', 'key': 's C',   'action': 'sort --add !leeches',
     'description': 'Sort trackers by number of downloading peers in reverse'},
    {'context': 'trackerlist', 'key': 's s',   'action': 'sort --add seeds',
     'description': 'Sort trackers by number of seeding peers'},
    {'context': 'trackerlist', 'key': 's S',   'action': 'sort --add !seeds',
     'description': 'Sort trackers by number of seeding peers in reverse'},
    {'context': 'trackerlist', 'key': 's o',   'action': 'sort --add domain',
     'description': 'Sort trackers by tracker\'s domain'},
    {'context': 'trackerlist', 'key': 's O',   'action': 'sort --add !domain',
     'description': 'Sort trackers by tracker\'s domain in reverse'},
    {'context': 'trackerlist', 'key': 's t',   'action': 'sort --add torrent',
     'description': 'Sort trackers by torrent name'},
    {'context': 'trackerlist', 'key': 's T',   'action': 'sort --add !torrent',
     'description': 'Sort trackers by torrent name in reverse'},
    {'context': 'trackerlist', 'key': 's e',   'action': 'sort --add error',
     'description': 'Sort trackers by error message'},
    {'context': 'trackerlist', 'key': 's E',   'action': 'sort --add !error',
     'description': 'Sort trackers by error message in reverse'},
    {'context': 'trackerlist', 'key': 's a',   'action': 'sort --add status',
     'description': 'Sort trackers by current status'},
    {'context': 'trackerlist', 'key': 's A',   'action': 'sort --add !status',
     'description': 'Sort trackers by current status in reverse'},
    {'context': 'trackerlist', 'key': 's ,',   'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'trackerlist', 'key': 's .',   'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'trackerlist', 'key': '/',     'action': ("interactive 'limit \"[]\"' --per-change "
                                                          "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed trackers by applying more filters'},

    # Tracker actions
    {'context': 'torrent', 'key': 'enter', 'action': 'tab summary',
     'description': 'Show details about torrent of focused tracker in new tab'},
    {'context': 'torrent', 'key': 'alt-enter', 'action': 'tab filelist',
     'description': 'List files of torrent of focused tracker in a new tab'},

    # Setting list actions
    {'context': 'settinglist', 'key': 's n', 'action': 'sort --add name',
     'description': 'Sort settings by name'},
    {'context': 'settinglist', 'key': 's N', 'action': 'sort --add !name',
     'description': 'Sort settings by name in reverse'},
    {'context': 'settinglist', 'key': 's v', 'action': 'sort --add value',
     'description': 'Sort settings by value'},
    {'context': 'settinglist', 'key': 's V', 'action': 'sort --add !value',
     'description': 'Sort settings by value in reverse'},
    {'context': 'settinglist', 'key': 's d', 'action': 'sort --add default',
     'description': 'Sort settings by default value'},
    {'context': 'settinglist', 'key': 's D', 'action': 'sort --add !default',
     'description': 'Sort settings by default value in reverse'},
    {'context': 'settinglist', 'key': 's c', 'action': 'sort --add description',
     'description': 'Sort settings by description'},
    {'context': 'settinglist', 'key': 's C', 'action': 'sort --add !description',
     'description': 'Sort settings by description in reverse'},
    {'context': 'settinglist', 'key': 's ,', 'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'settinglist', 'key': 's .', 'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'settinglist', 'key': '/',   'action': ("interactive 'limit \"[]\"' --per-change "
                                                        "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed settings by applying more filters'},

    # Setting actions
    {'context': 'setting', 'key': 'ctrl-r', 'action': 'reset',
     'description': 'Reset setting to its default value'},

    # Help actions
    {'context': 'helptext', 'key': '/',      'action': ("interactive 'find \"[]\"' --per-change "
                                                        "--on-cancel 'find --clear' --ignore-errors"),
     'description': 'Search for string'},
    {'context': 'helptext', 'key': 'n',      'action': 'find --next',
     'description': 'Find next occurence of search string'},
    {'context': 'helptext', 'key': 'ctrl-n', 'action': 'find --next',
     'description': 'Find next occurence of search string'},
    {'context': 'helptext', 'key': 'N',      'action': 'find --previous',
     'description': 'Find previous occurence of search string'},
    {'context': 'helptext', 'key': 'ctrl-p', 'action': 'find --previous',
     'description': 'Find previous occurence of search string'},
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


class CallbackDict(dict):
    """Mapping that runs a callback on every change"""
    # https://stackoverflow.com/a/5186698

    __slots__ = ["callback"]

    def __init__(self, callback, *args, **kwargs):
        self.callback = callback
        dict.__init__(self, *args, **kwargs)

    def _wrap(method):
        def wrapper(self, *args, **kwargs):
            result = method(self, *args, **kwargs)
            self.callback()
            return result
        return wrapper

    __delitem__ = _wrap(dict.__delitem__)
    __setitem__ = _wrap(dict.__setitem__)
    clear = _wrap(dict.clear)
    pop = _wrap(dict.pop)
    popitem = _wrap(dict.popitem)
    setdefault = _wrap(dict.setdefault)
    update = _wrap(dict.update)


from collections import abc
def listify_args(args):
    """
    Make list from `args`

    Ensure `args` is a string and use split(',') to turn it into a list.  If
    `args` is any non-string iterable, recursively listify each item and append
    it to the final list.

    Despite the name, return a tuple.
    """
    if args is None:
        return []

    if isinstance(args, str):
        # args is a string of comma-separated list items
        args = (str(arg).strip() for arg in args.split(','))
    elif isinstance(args, abc.Iterable):
        # args is an iterable of list items; each item may be a string of
        # comma-separated items
        args = (a for arg in args
                for a in listify_args(arg))
    else:
        # args is something else
        args = (str(args).strip())
    return [arg for arg in args if arg]


def log_msgs(process, response, quiet=False):
    """
    Log messages to `logger`

    `msgs` is an iterable of strings or exceptions.  Strings are logged to
    level INFO, exceptions are logged to level ERROR.

    If `quiet` evaluates to True, INFO messages are not logged.
    """
    if not quiet:
        for msg in response.msgs:
            process.info(msg)

    for error in response.errors:
        process.error(error)
